# -*- coding: utf-8 -*-
"""
Created on Fri Oct 21 10:43:59 2022

@author: chris.kerklaan
"""


import threedi_raster_edits as tre


raster_path_a = r"/path/to/raster.tif"

raster_path_b = r"/path/to/rasterb.tif"

raster_a = tre.Raster(raster_path_a)
raster_b = tre.Raster(raster_path_b)

help(raster_a.align)
raster_a.align(raster_b)



# -*- coding: utf-8 -*-
"""
Created on Fri Apr 26 14:39:17 2019

@author: chris.kerklaan

These are essentially deprecated. However we still need to find a way to make
this easily available from a single shapely object.

"""
from .point import Point, MultiPoint  # noqa
from .linestring import LineString, MultiLineString  # noqa
from .polygon import Polygon, MultiPolygon  # noqa



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 30 13:52:55 2019

@author: chris.kerklaan
"""
from threedi_edits.utils.project import Files
from threedi_edits.utils.project import Modules

files = Files(__file__)
modules = Modules(__name__)

# pyflakes
files
modules



# -*- coding: utf-8 -*-
"""
Created on Fri Jan 22 13:49:48 2021

@author: chris.kerklaan

Wrapper for an ogr geometry

#TODO:
    1. Take over the most important shapely functions.
    1. Convert/fix geometry should have an input geom_type and a target_geom_type
    
    3. Take over shapely's affine functions.
    4. Take over triangulate.
    5. Take over vonoroi shapely.ops.voronoi_diagram.
    6. Snapping.
    
"""
# First-party imports
import logging
import numpy as np

# Third-party imports
import shapely
import shapely.lib as shapely_lib
from osgeo import ogr, _ogr
from cached_property import cached_property

# Local imports
from .spatial_reference import SpatialReference

# GLOBALS
logger = logging.getLogger(__name__)

POLYGON = "POLYGON (({x1} {y1},{x2} {y1},{x2} {y2},{x1} {y2},{x1} {y1}))"
POINT = "POINT ({x1} {y1})"
LINESTRING = "LINESTRING ({x1} {y1}, {x2} {y2})"

TRANSLATION = {
    "point": ogr.wkbPoint,
    "line": ogr.wkbLineString,
    "polygon": ogr.wkbPolygon,
    "multipoint": ogr.wkbMultiPoint,
    "multiline": ogr.wkbMultiLineString,
    "multiPolygon": ogr.wkbMultiPolygon,
}

SINGLE_TYPES = {
    ogr.wkbMultiSurface: ogr.wkbPolygon,
    ogr.wkbMultiPolygon: ogr.wkbPolygon,
    ogr.wkbMultiLineString: ogr.wkbLineString,
    ogr.wkbMultiPoint: ogr.wkbPoint,
    ogr.wkbPolygon: ogr.wkbPolygon,
    ogr.wkbLineString: ogr.wkbLineString,
    ogr.wkbPoint: ogr.wkbPoint,
}

MULTI_TYPES = {
    ogr.wkbPoint: ogr.wkbMultiPoint,
    ogr.wkbLineString: ogr.wkbMultiLineString,
    ogr.wkbPolygon: ogr.wkbMultiPolygon,
    ogr.wkbMultiPoint: ogr.wkbMultiPoint,
    ogr.wkbMultiLineString: ogr.wkbMultiLineString,
    ogr.wkbMultiPolygon: ogr.wkbMultiPolygon,
}

# TODO Can be written easier, also does not cover everything.
ESTIMATES = {
    # multipolygons
    ogr.wkbMultiPolygon: ogr.wkbMultiPolygon,
    ogr.wkbMultiSurface: ogr.wkbMultiPolygon,
    ogr.wkbMultiSurfaceM: ogr.wkbMultiPolygon,
    ogr.wkbMultiSurfaceZ: ogr.wkbMultiPolygon,
    ogr.wkbMultiSurfaceZM: ogr.wkbMultiPolygon,
    ogr.wkbMultiPolygon25D: ogr.wkbMultiPolygon,
    ogr.wkbMultiPolygonM: ogr.wkbMultiPolygon,
    ogr.wkbMultiPolygonZM: ogr.wkbMultiPolygon,
    # polygons
    ogr.wkbPolygon: ogr.wkbPolygon,
    ogr.wkbCurvePolygon: ogr.wkbPolygon,
    ogr.wkbCurvePolygonM: ogr.wkbPolygon,
    ogr.wkbCurvePolygonZ: ogr.wkbPolygon,
    ogr.wkbCurvePolygonZM: ogr.wkbPolygon,
    ogr.wkbSurface: ogr.wkbPolygon,
    ogr.wkbSurfaceM: ogr.wkbPolygon,
    ogr.wkbSurfaceZ: ogr.wkbPolygon,
    ogr.wkbSurfaceZM: ogr.wkbPolygon,
    ogr.wkbPolygon25D: ogr.wkbPolygon,
    ogr.wkbPolygonM: ogr.wkbPolygon,
    ogr.wkbPolygonZM: ogr.wkbPolygon,
    # multilinestring
    ogr.wkbMultiLineString: ogr.wkbMultiLineString,
    ogr.wkbMultiCurve: ogr.wkbMultiLineString,
    ogr.wkbMultiCurveM: ogr.wkbMultiLineString,
    ogr.wkbMultiCurveZ: ogr.wkbMultiLineString,
    ogr.wkbMultiCurveZM: ogr.wkbMultiLineString,
    ogr.wkbMultiLineString25D: ogr.wkbMultiLineString,
    ogr.wkbMultiLineStringM: ogr.wkbMultiLineString,
    ogr.wkbMultiLineStringZM: ogr.wkbMultiLineString,
    # linestring
    ogr.wkbLineString: ogr.wkbLineString,
    ogr.wkbCircularString: ogr.wkbLineString,
    ogr.wkbCircularStringM: ogr.wkbLineString,
    ogr.wkbCircularStringZ: ogr.wkbLineString,
    ogr.wkbCircularStringZM: ogr.wkbLineString,
    ogr.wkbCurve: ogr.wkbLineString,
    ogr.wkbCurveM: ogr.wkbLineString,
    ogr.wkbCurveZ: ogr.wkbLineString,
    ogr.wkbCurveZM: ogr.wkbLineString,
    ogr.wkbLineString25D: ogr.wkbLineString,
    ogr.wkbLineStringM: ogr.wkbLineString,
    ogr.wkbLineStringZM: ogr.wkbLineString,
    # multipoints
    ogr.wkbMultiPoint: ogr.wkbMultiPoint,
    ogr.wkbMultiPoint25D: ogr.wkbMultiPoint,
    ogr.wkbMultiPointM: ogr.wkbMultiPoint,
    ogr.wkbMultiPointZM: ogr.wkbMultiPoint,
    # point
    ogr.wkbPoint: ogr.wkbPoint,
    ogr.wkbPoint25D: ogr.wkbPoint,
    ogr.wkbPointM: ogr.wkbPoint,
    ogr.wkbPointZM: ogr.wkbPoint,
}

SINGLE_TYPES_INVERT = dict(map(reversed, SINGLE_TYPES.items()))
SINGLE_TO_MULTIPLE = {1: [1, 4], 2: [2, 5], 3: [3, 6], 4: [4], 5: [5], 6: [6]}
ALLOWED = {1: [1, 4], 2: [2, 5], 3: [3, 6], 4: [4, 1], 5: [5, 2], 6: [6, 3]}
SHAPELY_INVALID = np.uint8(2)


class Geometry(ogr.Geometry):
    """
    The Geometry class is an inhertied version of an ogr geometry.

    Memory:
        The input is copied, hence the pointer is lost.
        The geometry in a feature must be set with set_feature.
        Every function returns a geometry.

    """

    def __init__(self, geometry: ogr.Geometry):
        super().__init__(wkb=convert_geometry(geometry).ExportToWkb())
        self.AssignSpatialReference(geometry.GetSpatialReference())

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        wkt = self.wkt
        max_length = 60
        if len(wkt) > max_length:
            return "<{}...>".format(wkt[: max_length - 3])

        return f"<{wkt}>"

    @classmethod
    def from_shapely(cls, geometry):
        return cls(ogr.CreateGeometryFromWkt(geometry.wkt))

    def check_type(self, types):
        if not self.type in types:
            raise TypeError(f"Wrong ogr geom type {self.name} {self.type_name}")

    @cached_property
    def shape(self):
        """laods a shapely geometry from wkt"""
        geometry = shapely.from_wkt(self.wkt)
        shapely.prepare(geometry)
        return geometry

    @property
    def area(self):
        return self.GetArea()

    @property
    def length(self):
        return self.Length()

    @property
    def name(self):
        return self.GetGeometryName()

    @property
    def type(self):
        return self.GetGeometryType()

    @property
    def is_multi(self):
        return self.type > 3

    @property
    def type_name(self):
        return ogr.GeometryTypeToName(self.type)

    @property
    def valid(self):
        return self.IsValid()

    @property
    def fixed(self):
        return fix(self)

    @property
    def points(self):
        return self.GetPoints()

    @property
    def envelope(self):
        return self.GetEnvelope()

    @property
    def wkt(self):
        return self.ExportToWkt()

    @property
    def wkb(self):
        return self.ExportToWkb()

    @property
    def wkb_size(self):
        return self.WkbSize()

    @property
    def point_on_surface(self):
        return Geometry(self.PointOnSurface())

    @property
    def centroid(self):
        return Geometry(self.Centroid())

    @property
    def boundary(self):
        return Geometry(self.GetBoundary())

    @property
    def spatial_reference(self):
        return SpatialReference.from_sr(self.GetSpatialReference(), self.centroid)

    @spatial_reference.setter
    def spatial_reference(self, sr_input):
        """takes an epsg or a spatial_reference input"""
        if type(sr_input) == int:
            sr_input = SpatialReference.from_epsg(sr_input)
        self.AssignSpatialReference(sr_input)

    @property
    def epsg(self):
        return self.spatial_reference.epsg

    @property
    def ogr(self):
        """returns a pure ogr geometry"""
        return release(self)

    def copy(self):
        return Geometry(geometry=self)

    def buffer(self, size):
        return Geometry(self.Buffer(size))

    def simplify(self, simplification):
        return Geometry(self.Simplify(simplification))

    def difference(self, geometry, ignore_errors=False):
        """returns a differenced geometry, or none if invalid or not intersects"""
        return Geometry(difference(self, geometry, ignore_errors))

    def clip(self, mask):
        """returns a list of clipped geometry in the correct output type
        or none
        """
        clipped = clip(self, mask, self.type)
        if clipped:
            return [Geometry(i) for i in clipped]
        else:
            return

    def reproject(self, out_epsg, use_points=False):
        """reprojects the geometries"""
        geometry = self.ogr
        geometry.Transform(self.spatial_reference.transform(out_epsg))
        return Geometry(geometry)

    def dissolve(self):
        """dissolved geometry, returns a single geometry"""
        return Geometry(dissolve(self), SINGLE_TYPES[self.type])

    def to_single(self):
        """creates single parts, returns a list or none"""
        return [Geometry(i) for i in multipart_to_singlepart(self)]

    def fix(self, geometry_type=None):
        return fix(self, geometry_type)

    def union(self, geometry):
        return Geometry(self.Union(geometry))

    def intersection(self, geometry):
        return Geometry(self.Intersection(geometry))


def shapely_geometry(geometry: ogr.Geometry) -> shapely.geometry.base.BaseGeometry:
    if geometry:
        geometry = convert_geometry(geometry)
        wkb = bytes(_ogr.Geometry_ExportToWkb(geometry))
        array = np.asarray(wkb, dtype=object)
        geometry = shapely_lib.from_wkb(array, SHAPELY_INVALID)
        shapely_lib.prepare(geometry)
        return geometry
    return None


def to_ogr(geometry):
    if isinstance(geometry, shapely.geometry.base.BaseGeometry):
        return ogr.CreateGeometryFromWkt(geometry.wkt)
    if hasattr(geometry, "ogr"):
        return geometry.ogr
    if isinstance(geometry, ogr.Geometry):
        return geometry
    raise ValueError("Cannot convert ogr geometry")


def to_shapely(geometry):
    if isinstance(geometry, ogr.Geometry):
        return shapely_geometry(geometry)

    elif hasattr(geometry, "ogr"):
        return shapely_geometry(geometry.ogr)

    elif hasattr(geometry, "shape"):
        return geometry.shape

    else:
        if isinstance(geometry, shapely.geometry.base.BaseGeometry):
            return geometry

    raise ValueError("Cannot convert to shapely geometry")


def release(geometry: ogr.Geometry):
    """releases the c pointer from the geometry"""
    output = ogr.CreateGeometryFromWkb(geometry.ExportToWkb())
    output.AssignSpatialReference(geometry.GetSpatialReference())
    return output


def estimate_geometry_type(geometry_type):
    """
    Estimates the target geometry type
    """
    if geometry_type > ogr.wkbUnknown and geometry_type < ogr.wkbMultiPolygon:
        return geometry_type

    if not geometry_type in ESTIMATES:
        name = ogr.GeometryTypeToName(geometry_type)
        raise ValueError(f"Cannot estimate geometry type for: {name}")

    return ESTIMATES[geometry_type]


def convert_geometry(geometry, target_type=None) -> ogr.Geometry:
    """
    We only use planar and linear geometries.
    Therefore we convert the geometry to the target type.
    if the target type is not present we estimate it.

    We account for the following cases:
        - geometry type is 3D.
        - geometry type is curved.
        - geometry type is a collecton.

    returns a single, linear and planar geometry, or raises an error.


    """
    geometry_type = geometry.GetGeometryType()

    if target_type is None:
        target_type = estimate_geometry_type(geometry_type)

    if geometry_type == target_type:
        return geometry

    # 3D conversion
    if geometry.Is3D():
        geometry.FlattenTo2D()

    # Curved geometrya
    if geometry.HasCurveGeometry():
        geometry = geometry.GetLinearGeometry()

    # Geometry type, also looks for Z, M and ZM.
    if str(ogr.wkbGeometryCollection) in str(geometry_type):
        logger.debug("Found a geometry collection, retrieving only the first geometry")
        for part in geometry:
            if part.GetGeometryType() == target_type:
                geometry = part
                break

    # Finally, if all of the above does not work, we try to force.
    geometry = ogr.ForceTo(geometry, target_type)

    return geometry


def fix(geometry, target_geometry_type=None):
    """
    Fixes a geometry is it is invalid or has the wrong geometry type.
    Fixes of the wrong type using convert_geometry.
    Things that we fix:
        - Point count of 1 in a linestring returns an error.
        - Invalid polygons of multipolygons receive a buffer.

    Everything else is done with make valid.


    """
    if geometry is None:
        return None

    geometry_type = geometry.GetGeometryType()
    geometry_name = ogr.GeometryTypeToName(geometry_type)

    if not target_geometry_type:
        target_geometry_type = estimate_geometry_type(geometry_type)

    geometry = convert_geometry(geometry, target_geometry_type)

    if geometry.IsValid():
        return geometry

    if geometry_type == ogr.wkbLineString:
        if geometry.GetPointCount() == 1:
            logger.debug("Geometry point count of linestring = 1")
            return geometry, False

    if geometry_type in [ogr.wkbPolygon, ogr.wkbMultiPolygon]:
        logger.debug("Buffered the geometry")
        geometry = geometry.Buffer(0)

    if not geometry.IsValid():
        logger.deug("Using MakeValid")
        geometry = geometry.MakeValid()  # multipolygon to collection.

    if not geometry.IsValid():
        logger.warning("Failed to fix the geometry, returning None.")
        return None

    if not geometry.GetGeometryType() == target_geometry_type:
        target_name = ogr.GeometryTypeToName(target_geometry_type)
        logger.warning(f"Could not convert {geometry_name} to {target_name}.")

    return geometry


def multipart_to_singlepart(geometry: ogr.Geometry):
    """returns a list of geometries, if invalid none"""
    if "MULTI" in geometry.GetGeometryName():
        return [release(part) for part in geometry]
    else:
        return [geometry]


def clip(geometry: ogr.Geometry, mask: ogr.Geometry, output_type: int):
    """Clips a single geometry on a mask
    returns invalid geoms as none
    returns a geometry of an allowed output type
    returns a list
    """
    within = geometry.Within(mask)

    # check for the boundary
    if geometry.Intersects(mask) and not within:
        intersect = geometry.Intersection(mask)
        intersect_type = intersect.GetGeometryType()
        if intersect_type not in ALLOWED[output_type]:
            logger.debug("Output has incorrect geometry type")
            return

        if intersect_type > 3:
            return [intersect]
        else:
            return multipart_to_singlepart(intersect)

    elif within:
        return [geometry]

    else:
        logger.debug("Outside of geometry")
        return  # outside


def difference(geometry: ogr.Geometry, mask: ogr.Geometry, ignore_errors=False):
    """checks if intersects before doing difference"""

    try:
        if geometry.Intersects(mask):
            return geometry.Difference(mask)
    except RuntimeError:
        if not ignore_errors:
            raise RuntimeError("Difference")
        logger.debug("Geometry difference runtime error - ignore = True")

    # if not intersects or ignore errors return original geometry
    return geometry


def dissolve(multigeometry: ogr.Geometry):
    """dissolves a multigeometry into a single geometry"""
    geometry_type = multigeometry.GetGeometryType()

    if geometry_type == ogr.wkbMultiPolygon:
        union = multigeometry.UnionCascaded()
    else:
        union = ogr.Geometry(SINGLE_TYPES[geometry_type])
        for geometry in multigeometry:
            union = union.Union(geometry)

    return union



# -*- coding: utf-8 -*-
"""
Created on Wed May 25 09:03:46 2022

@author: stijn.overmeen

Converts channel profiles to threedi cross section definitions and locations
Add these to the threedi model
Create vertex in channel

Bonus:
    Search for unused cross section defintions and delete these
"""

import os
import pandas as pd
from pathlib import Path
import matplotlib.pyplot as plt

plt.ioff()  # don't show plots in console
import numpy as np
import sys

sys.path.append(r"C:/Users/stijn.overmeen/Documents/GitHub/threedi-raster-edits")
import threedi_edits as tre
from osgeo import ogr, osr
from threedi_edits.threedi.tables.templates import Templates

template = Templates("integrated")

# packages for logging
import datetime, time, logging

# Define working dir
FOLDER = Path(
    "C:/Users/stijn.overmeen/Documents/Projecten_lokaal/Extern/overijssels_kanaal/scripts"
)

# Logging
LOG_FOLDER = os.path.join(FOLDER, "log")
if not os.path.exists(LOG_FOLDER):
    os.makedirs(LOG_FOLDER)
lines = ["Profielen Overijssels kanaal to threedi", ""]
log = (
    f"{datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d__%H-%M-%S')}"
    ".log"
)
with open(os.path.join(FOLDER, "log", log), "w") as f:
    for line in lines:
        f.write(line)
        f.write("\n")

logging.basicConfig(
    level=logging.INFO,
    handlers=[
        logging.FileHandler(os.path.join(FOLDER, "log", log)),
        logging.StreamHandler(),
    ],
    format="%(message)s",
    force=True,
)


def strings_to_floats(data):
    """
    converts strings to floats in the dataframe (specific actions per column name)
    """
    for col in columns_to_convert:
        if col in ["Afstand", "z"]:
            data[col] = pd.to_numeric(data[col].str.replace(",", "."))
        elif col in ["X", "Y"]:
            temp = data[col].str.replace(".", "")
            for i, v in temp.items():
                temp[i] = float(v[:6] + "." + v[6:])
            data[col] = pd.to_numeric(temp)
    return data


def yz_array_to_threedi_tabulated_definitions(yz_array, step_size):
    """
    convert yz array to threedi tabulated trapezium definitions
    threedi requires symmetrical profiles and continuously vertically increasing volume
    calculates flow profile for every vertical step and adapts definition accordingly
    """
    bottom_level = min(yz_array[:, 1])
    ind_middle = np.argmin(yz_array[:, 1])
    left_bank_level = max(yz_array[: ind_middle + 1, 1])
    right_bank_level = max(yz_array[ind_middle:, 1])
    left_bank_above_bottom = left_bank_level - bottom_level
    right_bank_above_bottom = right_bank_level - bottom_level
    left_bank = min(yz_array[:, 0])
    yz_array[:, 1] = yz_array[:, 1] - bottom_level
    yz_array[:, 0] = yz_array[:, 0] - left_bank
    min_z = min(yz_array[:, 1])
    max_z = min(left_bank_above_bottom, right_bank_above_bottom)  # max(yz_array[:,1])
    steps = np.arange(min_z, max_z, step_size) + step_size
    channel_area_under_step = np.empty(steps.shape)
    width = np.empty(steps.shape)
    for i, step in enumerate(steps):
        y = yz_array[:, 0]
        z = yz_array[:, 1]
        z_clip = np.clip(z, 0, step)
        area_under_clip = np.trapz(z_clip, y, dx=0.01)
        area_under_step = step * (max(y) - min(y))
        channel_area_under_step[i] = area_under_step - area_under_clip
    area_increase = channel_area_under_step[1:] - channel_area_under_step[:-1]
    width = np.insert(area_increase / step_size, 0, 0, axis=0)
    height = np.linspace(0, max_z, width.shape[0])
    middle = max(width) / 2
    yz_tab_list = []
    for h, w in zip(height, width):
        if h == 0:
            yz_tab_list.append([middle, h])
        else:
            z = h
            yl = middle - w / 2
            yr = middle + w / 2
            yz_tab_list.append([yl, z])
            yz_tab_list.append([yr, z])
    yz_tab = np.array(yz_tab_list)
    yz_tab_left = yz_tab[yz_tab[:, 0] < middle]
    yz_tab_right = yz_tab[yz_tab[:, 0] >= middle]
    yz_tab_left_sorted = yz_tab_left[
        np.lexsort((yz_tab_left[:, 0], -yz_tab_left[:, 1]))
    ]
    yz_tab_right_sorted = yz_tab_right[
        np.lexsort((yz_tab_right[:, 0], yz_tab_right[:, 1]))
    ]
    yz_tab_sorted = np.concatenate((yz_tab_left_sorted, yz_tab_right_sorted), axis=0)
    np.set_printoptions(precision=2)
    widthstring = np.array2string(
        width, formatter={"float_kind": lambda width: "%.2f" % width}
    )
    heightstring = np.array2string(
        height, formatter={"float_kind": lambda height: "%.2f" % height}
    )

    return (
        left_bank_level,
        right_bank_level,
        widthstring,
        heightstring,
        bottom_level,
        yz_tab_sorted,
    )


def plot_profile_comparisons(profile_name, yz_tab_sorted, yz_array):
    """
    plots and save figures of profile comparisons, original yz against threedi tabulated definition
    """
    plt.figure(figsize=(10, 5))
    plt.title(
        "Comparison between YZ and tabulated trapezium definitions for "
        f"{profile_name}"
    )
    plt.plot(yz_tab_sorted[:, 0], yz_tab_sorted[:, 1], label="Tabulated trapezium")
    plt.plot(yz_array[:, 0], yz_array[:, 1], label="YZ")
    plt.xlabel("Distance from bank [m]", fontdict=None, labelpad=None)
    plt.ylabel("Depth [m]", fontdict=None, labelpad=None)
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.0)
    plt.savefig(
        os.path.join(FOLDER, "profiel_figuren", "slib", f"{profile_name}"),
        bbox_inches="tight",
    )


def store_information_cross_sections_in_dict(
    profile_name,
    profile_df,
    left_bank_level,
    right_bank_level,
    widthstring,
    heightstring,
    bottom_level,
):
    """
    returns dict with relevant information for further processing of information to 3Di model
    """
    cross_section = {}
    cross_section["code"] = profile_name
    cross_section["height"] = heightstring
    cross_section["width"] = widthstring
    cross_section["bank_level"] = min(left_bank_level, right_bank_level)
    cross_section["code"] = profile_name
    cross_section["reference_level"] = bottom_level
    cross_section["X"] = profile_df["X"].iloc[0]
    cross_section["Y"] = profile_df["Y"].iloc[0]

    return cross_section


def add_cross_section_definition_to_model(model, cs_def_profile, shape):
    """
    add cross section definition

    needed:
        code, width string, height string from profile defintion dict
        shape (integer)
    """
    cs_def_threedi = template.cross_section
    cs_def_threedi["code"] = cs_def_profile["code"]
    cs_def_threedi["shape"] = shape  # tabulated trapezium
    cs_def_threedi["width"] = cs_def_profile["width"].replace("[", "").replace("]", "")
    cs_def_threedi["height"] = (
        cs_def_profile["height"].replace("[", "").replace("]", "")
    )
    model.cross_section_definitions.add(cs_def_threedi)

    return model


def add_cross_section_location_to_model(
    model, cs_loc_profile, friction_type, friction_value, EPSG_number
):
    """
    add cross section location, geometry based on X, Y coordinates
    finds closest channel based on spatial search, and adds this channel id to the table

    needed:
        code, reference level, bank level, X and Y coordinates from profile defintion dict
        EPSG code beloging to X and Y coordinates
        friction type and friction value
    """
    cs_loc_threedi = template.cross_section_location
    cs_loc_threedi["code"] = cs_loc_profile["code"]
    # definition id based on profile code
    for cs_def_threedi in subselection_of_cs_definitions:
        if cs_def_threedi["code"] == cs_loc_profile["code"]:
            cs_loc_threedi["definition_id"] = cs_def_threedi.fid
            break
    cs_loc_threedi["reference_level"] = cs_loc_profile["reference_level"]
    cs_loc_threedi["bank_level"] = cs_loc_profile["bank_level"]
    cs_loc_threedi["friction_type"] = friction_type
    cs_loc_threedi["friction_value"] = friction_value

    # TODO
    # Tip!
    # zo zou ik het doen:
    point = tre.Point.from_point(
        (float(cs_loc_profile["X"]), float(cs_loc_profile["Y"])), epsg=28992
    )
    point = point.reproject(4326)

    point = ogr.Geometry(ogr.wkbPoint)
    point.AddPoint(float(cs_loc_profile["X"]), float(cs_loc_profile["Y"]))

    # transform point
    source = osr.SpatialReference()
    source.ImportFromEPSG(EPSG_number)
    target = osr.SpatialReference()
    target.ImportFromEPSG(4326)

    # The line below sets the TRADITIONAL_GIS_ORDER.
    target.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

    transform = osr.CoordinateTransformation(source, target)
    point.Transform(transform)

    cs_loc_threedi["the_geom"] = tre.gis.point.Point(point)

    # TODO
    # Tip!, je kunt spatial filteren op channels. Ik kan me voorstellen dat de onderste code wat traag is.
    # Zo zou ik het doen met een prefilter:
    convert = tre.threedi.utils.conversions.METER_TO_DEGREE
    buffer = convert * 100  # meter
    output = []
    while len(output) == 0:
        output = model.channels.spatial_filter(point.Buffer(buffer), return_vector=True)
        buffer = buffer + convert * 100  # extra 100 meter zoekafstand

    # find closest channel
    smallest_distance = 99999999  # initialise big
    for channel in output:
        distance = point.Distance(channel.geometry)
        if distance < smallest_distance:
            smallest_distance = distance
            closest_channel = channel

    # find closest channel
    smallest_distance = 99999999  # initialise big
    for channel in model.channels:
        distance = point.Distance(channel.geometry)
        if distance < smallest_distance:
            smallest_distance = distance
            closest_channel = channel

    # add channel_id to cross section location
    cs_loc_threedi["channel_id"] = closest_channel.fid
    model.cross_section_locations.add(cs_loc_threedi)

    return model, closest_channel, point


if __name__ == "__main__":

    file_hoofdtak = Path(
        os.path.join(FOLDER, "data", "profielen_ovkanaal (bewerking2)_afvoer_slib.csv")
    )  # hoofdkanaal profielen

    file_zijtak_profiles = Path(
        os.path.join(FOLDER, "data", "inmeting 2020 - lettele.xlsx")
    )  # zijtak profielen
    file_zijtak_coord = Path(
        os.path.join(FOLDER, "data", "PROFILES_Locations.csv")
    )  # bijbehorende coordinaten

    hoofdtak_df = pd.read_csv(file_hoofdtak, sep=";", decimal=",")  # read file

    columns_to_convert = ["Afstand", "z", "X", "Y"]
    hoofdtak_df_improved = strings_to_floats(hoofdtak_df)

    zijtak_profiles_df = pd.read_excel(
        file_zijtak_profiles, sheet_name="bagger", decimal=","
    )  # read file
    zijtak_coord_df = pd.read_csv(file_zijtak_coord, sep=";", decimal=".")

    rename_dict_coord = {"ID": "profielnaam", "x": "X", "y": "Y"}

    rename_dict_profiles = {
        "Profiel": "profielnaam",
        "x ": "Afstand",
        "Type": "type",
        "Bron": "bron",
    }

    zijtak_profiles_df.rename(columns=rename_dict_profiles, inplace=True)
    zijtak_coord_df.rename(columns=rename_dict_coord, inplace=True)

    zijtak_df = zijtak_profiles_df.merge(zijtak_coord_df, on="profielnaam")

    data = pd.concat([hoofdtak_df_improved, zijtak_df], ignore_index=True)

    yz_profiles = []  # create empty list for YZ-profiles
    cross_sections = []  # create empty list for storing cross section information
    profiles = data.groupby(data["profielnaam"])  # group data by profile name

    logging.info("converting profiles to threedi tabulated definitions...")

    # TODO
    # Tip!
    # ik zou loopen met tre.Progress(profiles, "converting profiles to threedi tabulated defintions")
    for profile in profiles:

        profile_name = profile[0]
        logging.info(profile_name)
        profile_df = profile[1]
        yz_array = profile_df[["Afstand", "z"]].to_numpy()

        (
            left_bank_level,
            right_bank_level,
            widthstring,
            heightstring,
            bottom_level,
            yz_tab_sorted,
        ) = yz_array_to_threedi_tabulated_definitions(yz_array=yz_array, step_size=0.1)
        # plot_profile_comparisons(profile_name,yz_tab_sorted,yz_array)

        cross_section = store_information_cross_sections_in_dict(
            profile_name=profile_name,
            profile_df=profile_df,
            left_bank_level=left_bank_level,
            right_bank_level=right_bank_level,
            widthstring=widthstring,
            heightstring=heightstring,
            bottom_level=bottom_level,
        )
        cross_sections.append(cross_section)

    sqlite_path = r"C:/Users/stijn.overmeen/Documents/Projecten_lokaal/Extern/overijssels_kanaal/speeltuin/zandkasteel.sqlite"
    model = tre.ThreediEdits(sqlite_path, mode="memory")

    logging.info("\nadding cross section definitions...")
    codes = []  # appending relevant codes for later use in for-loop
    for cs_def_profile in cross_sections:
        codes.append(cs_def_profile["code"])
        model = add_cross_section_definition_to_model(
            model=model, cs_def_profile=cs_def_profile, shape=6
        )
        logging.info(f"added cross section definition for {cs_def_profile['code']}")

    # get subselection of cross section definitions you just added for looping over while adding cross section locations
    subselection_of_cs_definitions = []

    for cs_def_threedi in model.cross_section_definitions:
        if cs_def_threedi["code"] in codes:
            subselection_of_cs_definitions.append(cs_def_threedi)

    logging.info("\nadding cross section locations...")
    for cs_loc_profile in cross_sections:
        logging.info("\n" + cs_loc_profile["code"])
        model, closest_channel, point = add_cross_section_location_to_model(
            model=model,
            cs_loc_profile=cs_loc_profile,
            friction_type=2,
            friction_value=0.033,
            EPSG_number=28992,
        )
        logging.info("added cross section location")

        # delete existing cross section locations of this channel, but not the ones we might have added previously
        # so search for code name starting with any of the following prefixes, keep these
        prefixes = ["prof_D2018", "AS1", "AS2"]
        for cs_loc_threedi_for_deleting_old_ones in model.cross_section_locations:
            if (
                cs_loc_threedi_for_deleting_old_ones["channel_id"]
                == closest_channel.fid
            ):
                if not cs_loc_threedi_for_deleting_old_ones["code"].startswith(
                    tuple(prefixes)
                ):
                    model.cross_section_locations.delete(
                        cs_loc_threedi_for_deleting_old_ones
                    )
                    logging.info(
                        "deleted old cross section location "
                        + cs_loc_threedi_for_deleting_old_ones["code"]
                    )

        logging.info("done, no old cross section locations left to delete")

        # add vertex to channel
        closest_channel.geometry = closest_channel.geometry.add_vertice(
            tre.gis.point.Point(point)
        )
        logging.info(
            f"added vertex to channel {closest_channel.fid} of cross section location"
        )

    # delete unused cross section definitions
    # sample used
    used_cs_def_ids = []
    for cs_loc in model.cross_section_locations:
        used_cs_def_ids.append(cs_loc["definition_id"])
    for orifice in model.orifices:
        used_cs_def_ids.append(orifice["cross_section_definition_id"])
    for culvert in model.culverts:
        used_cs_def_ids.append(culvert["cross_section_definition_id"])
    for weir in model.weirs:
        used_cs_def_ids.append(weir["cross_section_definition_id"])

    for cs_df in model.cross_section_definitions:
        if not cs_df.fid in used_cs_def_ids:  # delete if not used
            model.cross_section_definitions.delete(cs_df)

    # write features
    # model.cross_section_locations.write("cs_loc_test.gpkg")
    # model.channels.write("channels_test.gpkg")

    # write model
    model.write(
        sqlite_path.replace("zandkasteel", "model_met_ingemeten_profielen_slib"),
        check=True,
        rasters=True,
    )



# -*- coding: utf-8 -*-
"""
Created on Mon Jul 25 15:13:41 2022

@author: chris.kerklaan

This scripts creates a plot along a model of a channel.

For connection node the results ore sampled on a waterlevel, then the watersampled are put on  aline.



1. Create a view 


"""

from threedigrid.admin.gridresultadmin import GridH5ResultAdmin
from threedigrid.admin.gridadmin import GridH5Admin
import threedi_edits as tre
import numpy as np
import matplotlib.pyplot as plt

# globals
MIN_LEVEE_HOOGTE = 6.31


class Grid:
    def __init__(self, results_folder, start_time=0, end_time=None):
        self.h5_file_path = results_folder + "/gridadmin.h5"
        self.netcdf_file_path = results_folder + "/results_3di.nc"
        self.gr = GridH5ResultAdmin(self.h5_file_path, self.netcdf_file_path)
        self.ga = GridH5Admin(self.h5_file_path)

        self.start_time = start_time
        self.end_time = end_time
        if not end_time:
            self.end_time = self.gr.nodes.timestamps[-1]

    @property
    def nodes(self):
        if not hasattr(self, "_nodes"):
            self.set_nodes()

        return self._nodes

    def set_nodes(self):
        self._nodes = tre.Vector.from_scratch("nodes", 1, 28992)
        coordinates = np.vstack(self.gr.nodes.coordinates.T)
        ids = self.gr.nodes.id
        for id, coordinate in zip(ids, coordinates):
            self._nodes.add(geometry=tre.Point.from_point(coordinate), fid=id)

    def level(self, grid_id=[], model_id=[], start_time=0, end_time=None):
        """retrieves the waterlevel for different either model ids or grid ids
        params:
            grid_id: grid id, list or integer
            model_id: model id, list or integer
            start_time: start time, as a datetime object
            end_time: end time, as a datetime object
        """

        if end_time:
            self.end_time = end_time

        if type(grid_id) == int:
            return list(
                self.gr.nodes.filter(id=grid_id)
                .timeseries(self.start_time, self.end_time)
                .s1.flatten()
            )

        if type(model_id) == int:
            return list(
                self.gr.nodes.filter(content_pk=model_id)
                .timeseries(self.start_time, self.end_time)
                .s1.flatten()
            )

        output_dict = {}
        if len(grid_id) > 0:
            for gid in grid_id:
                output_dict[gid] = list(
                    self.gr.nodes.filter(id=gid)
                    .timeseries(self.start_time, self.end_time)
                    .s1.flatten()
                )

        if len(model_id) > 0:
            for mid in model_id:
                output_dict[mid] = list(
                    self.gr.nodes.filter(content_pk=mid)
                    .timeseries(self.start_time, self.end_time)
                    .s1.flatten()
                )

        return output_dict


def create(
    subject,
    result_folder,
    schematisation_path,
    selection_shape,
    output_name,
    start_time=43200,
    start_node_id=74507,
):

    results = Grid(result_folder, start_time=start_time, end_time=start_time + 350)

    # First select connection nodes with shape
    shape = tre.Vector(selection_shape).reproject(4326)[0].geometry

    # Open schematisation
    schema = tre.ThreediEdits(schematisation_path)

    node_fids = schema.nodes.spatial_filter(shape, "Within", return_fid=True)

    output = {"nodes": node_fids, "waterstand": [], "oeverhoogte": []}
    levels = results.level(model_id=node_fids)

    for node_id, level in levels.items():
        output["waterstand"].append(level[0])

    # now we find the bank levels
    channel = schema.channels.spatial_filter(shape, "Within", return_vector=True)

    # reproject
    channel = channel.copy().reproject(28992)
    levees = schema.levees.copy().reproject(28992)

    for node_id in node_fids:
        buffered_node = schema.nodes[node_id].geometry.buffer(0.00001)
        buffered_node = buffered_node.reproject(28992)
        clipped_channel = channel.clip(geometry=buffered_node).dissolve(quiet=True)
        line = clipped_channel[0]

        if line.geometry.IsEmpty():
            output["oeverhoogte"].append([None, None])
            continue

        perpendicular_line = line.geometry.perpendicular_lines(0.5, 20)[0]
        levee = levees.spatial_filter(perpendicular_line, return_vector=True)
        if levee.count > 0:
            if levee.count == 2:
                output["oeverhoogte"].append(
                    [
                        levee[levee.fids[0]]["crest_level"],
                        levee[levee.fids[1]]["crest_level"],
                    ]
                )
                print(levee.fids, node_id)
            else:
                output["oeverhoogte"].append(
                    [levee[levee.fids[0]]["crest_level"], None]
                )

        else:
            output["oeverhoogte"].append([None, None])

    # reorder from distance to gemaal
    start_node = schema.nodes[start_node_id].geometry

    distances = []
    for node_id in output["nodes"]:
        distances.append((schema.nodes[node_id].geometry.Distance(start_node), node_id))

    ordered_output = {"nodes": [], "waterstand": [], "oeverhoogte": [], "afstand": []}
    for (d, node_id) in sorted(distances):
        index = output["nodes"].index(node_id)
        ordered_output["nodes"].append(str(node_id))
        ordered_output["oeverhoogte"].append(output["oeverhoogte"][index])
        if output["waterstand"][index] is not None:
            ordered_output["waterstand"].append(output["waterstand"][index])
        else:
            ordered_output["waterstand"].append(output["waterstand"][index])

        ordered_output["afstand"].append(d / 0.00001 / 1000)

    # create data
    # delete nodata
    deletes = []
    for i, ws in enumerate(ordered_output["waterstand"]):
        if ws == -9999:
            deletes.append(i)

    for did in deletes:
        del ordered_output["nodes"][did]
        del ordered_output["waterstand"][did]
        del ordered_output["oeverhoogte"][did]
        del ordered_output["afstand"][did]

    oeverhoogtes = []
    for i in ordered_output["oeverhoogte"]:
        for hoogte in i:
            if hoogte is not None:
                oeverhoogtes.append(hoogte)

    minimale_oeverhoogte = min(oeverhoogtes)

    # plot lines
    fig, axs = plt.subplots(2, figsize=(10, 10))

    fig.suptitle(f"Verhanglijn en oeverhoogtes van het Overijssels Kanaal {subject}")

    axs[0].plot(
        ordered_output["afstand"],
        ordered_output["waterstand"],
        label=f"Waterstand na {int(start_time/3600)} uur, ~{int(start_time/3600/24)} dagen",
    )
    axs[0].plot(
        ordered_output["afstand"],
        [minimale_oeverhoogte for i in range(len(ordered_output["afstand"]))],
        label="Laagste oeverhoogte",
    )

    axs[0].set_xlabel("Afstand vanaf gemaal Ankersmit [kilometer] ")
    axs[0].set_ylabel("Waterstand [mNAP]")

    axs[0].scatter(
        ordered_output["afstand"],
        [i[0] for i in ordered_output["oeverhoogte"]],
        color="red",
        label="Oeverhoogte Zuid Oost",
        s=5,
    )

    axs[0].scatter(
        ordered_output["afstand"],
        [i[1] for i in ordered_output["oeverhoogte"]],
        color="blue",
        label="Oeverhoogte Noord West",
        s=5,
    )
    axs[0].legend()
    axs[0].grid()

    axs[1].plot(
        ordered_output["afstand"],
        ordered_output["waterstand"],
        label=f"Waterstand na {int(start_time/3600)} uur, ~{int(start_time/3600/24)} dagen ",
    )
    axs[1].set_xlabel("Afstand vanaf gemaal Ankersmit [kilometer] ")
    axs[1].set_ylabel("Waterstand [mNAP]")

    axs[1].legend()
    axs[1].grid()

    fig.show()
    fig.savefig(output_name, dpi=400)


if __name__ == "__main__":
    schematisation_path = r"C:\Users\chris.kerklaan\Documents\threedi_models\NBW Overijssels Kanaal - winter GHG - NVO\work in progress\schematisation/NBW Overijssels Kanaal - winter GHG - NVO.sqlite"
    result_folder = r"C:\Users\chris.kerklaan\Documents\threedi_models\NBW Overijssels Kanaal - winter GHG - NVO\work in progress\results\sim_86859_T100_winter_GHG_-_NVO_(n_=_0.3)"
    selection_shape = (
        r"C:/Users/chris.kerklaan/Documents/Projecten/overijssels_kanaal/data/ovk.shp"
    )
    output_name = r"C:\Users\chris.kerklaan\Documents\Projecten\overijssels_kanaal\processing/sideview_NVO_manning_03.png"
    create(
        "NVO (Manning n=0.3)",
        result_folder,
        schematisation_path,
        selection_shape,
        output_name,
        start_time=120 * 3600,
        start_node_id=74507,
    )



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 10 13:05:21 2021

@author: chris.kerklaan

Note:
    Spatial references can be either projected, geographic or local
    
"""

import logging
import osgeo
from osgeo import osr

# Logger
logger = logging.getLogger(__name__)

# Globals
STANDARD_EPSG = 28992


class SpatialReference(osr.SpatialReference):
    def __init__(self, wkt=None, centroid=None):

        if not wkt:
            if centroid:
                epsg = centroid_to_epsg(centroid)
                logger.debug(f"No spatial reference assigned, assigning epsg {epsg}")
                wkt = self.epsg_to_wkt(epsg)
            else:
                logger.debug("No spatial reference assigned, pleas asign epsg")

        if type(wkt) == int:
            logger.debug("please use SpatialReference.from_epsg")
        elif type(wkt) == osr.SpatialReference:
            logger.debug("please use SpatialReference.from_sr")

        super().__init__(wkt=wkt)

    @classmethod
    def from_epsg(cls, epsg, **kwargs):
        return cls(SpatialReference.epsg_to_wkt(epsg), **kwargs)

    @classmethod
    def from_sr(cls, sr, centroid=None):
        if sr == None:
            return cls(None, centroid)
        else:
            return cls(sr.ExportToWkt())

    @staticmethod
    def epsg_to_wkt(epsg):
        sr = osr.SpatialReference()
        sr.ImportFromEPSG(epsg)
        return sr.ExportToWkt()

    def __getitem__(self, item):
        return self.GetAttrValue(item)

    def __str__(self):
        return self.ExportToPrettyWkt()

    def __repr__(self):
        return f"EPSG:{self.epsg}, UNIT:{self.unit}"

    @property
    def wkt(self):
        return self.ExportToWkt()

    @property
    def unit(self):
        return self["UNIT"]

    @property
    def type(self):
        if self.IsProjected:
            return "PROJCS"
        elif self.IsGeographic:
            return "GEOGCS"
        elif self.IsLocal:
            return "LOCAL"
        else:
            return None

    @property
    def original(self):
        return osr.SpatialReference(self.wkt)

    @property
    def epsg(self):
        if self.GetAuthorityName(self.type) == "EPSG":
            epsg = self.GetAuthorityCode(self.type)
        else:
            self.AutoIdentifyEPSG()
            epsg = self.GetAttrValue("AUTHORITY", 1)

        # override amersfoort # in gdal 2 this is sometimes
        # not found correctly. utr-rast-task-01
        if ("Amersfoort" in self.wkt) and epsg is None:
            epsg = 28992

        if epsg == None:
            logger.debug(f"Could not find epsg, using {STANDARD_EPSG}")
            print(f"Could not find epsg, using {STANDARD_EPSG}")
            epsg = STANDARD_EPSG

        return int(epsg)

    def transform(self, epsg=28992):
        """returns a transform between current projection and out_epsg"""
        new_sr = SpatialReference.from_epsg(epsg).original
        sr = self.original

        if int(osgeo.__version__[0]) >= 3:
            # GDAL 3 axis order: https://github.com/OSGeo/gdal/issues/1546
            sr.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)
            new_sr.SetAxisMappingStrategy(osgeo.osr.OAMS_TRADITIONAL_GIS_ORDER)

        return osr.CoordinateTransformation(sr, new_sr)


def centroid_to_epsg(centroid):
    """estimating based on centroid coordinatess"""
    x, y = centroid.points[0]

    if -90 > x > 90 and -180 > y > 180:
        return 4326

    return 28992



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 10 15:08:00 2021

@author: chris.kerklaan
"""
# Import the Vector and Raster classes from threedi_edits
from threedi_raster_edtis import Vector, Raster
import numpy as np

# path to data
path_to_raster = "path/to/raster.tif"
path_to_vector = "path/to/vector.shp"
output_path = "path/to/output.shp"

# load raster and vector data
vector = Vector(path_to_vector)
raster = Raster(path_to_raster)

# copy the vector without the data
output = vector.copy(shell=True)

# add a field to the output
output.add_field("sum", float)

# loop over each feature in the vector
for feature in vector:

    # read the raster data
    raster_data = raster.read(feature.geometry)

    # do something with the data
    data_sum = np.nansum(raster_data)

    # write the output
    items = {"sum": data_sum}
    output.add(items=items, geometry=feature.geometry)


# Write to output_path
output.write(output_path)



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 15 08:58:57 2021

@author: chris.kerklaan

We wrap a feature for threedimodels here.

"""
# system imports
import logging

# Third-party import
import warnings
from osgeo import ogr

# Local imports
from threedi_edits.gis.feature import Feature
from threedi_edits.gis.fields import OGR_TYPES_INVERT, Fields
from threedi_edits.threedi.utils.exceptions import ThreediValueTypeError
from threedi_edits.threedi.utils.exceptions import ThreediValueTypeWarning
from threedi_edits.threedi.utils.exceptions import _HAPPY
from threedi_edits.threedi.utils.exceptions import custom_formatwarning

from threedi_edits.globals import SUPPORTED_THREEDI_VERSIONS
from threedi_edits.threedi.constants import get_version

V = get_version(SUPPORTED_THREEDI_VERSIONS[0])

# Globals
PROPERTIES = V.Properties()
warnings.formatwarning = custom_formatwarning


logger = logging.getLogger(__name__)


class ThreediFeature(Feature):

    instance = "threedi.feature.ThreediFeature"

    def __init__(
        self,
        feature: ogr.Feature,
        layer: ogr.Layer,
        fields: Fields,
        table_name: str,
        custom_set_function=None,
    ):
        Feature.__init__(self, feature, layer, fields)
        self.table_name = table_name
        self.custom_set_function = custom_set_function
        self.properties = PROPERTIES[self.table_name]

    def __str__(self):
        return f"3Di row, items: {self.all_items}"

    def __repr__(self):
        return f"({self.instance}) 3Di row, items: {self.all_items}"

    def __getitem__(self, field_name):
        return self.correct_type(field_name, self.get_item(field_name))

    def __setitem__(self, field_name, value):
        """setitem uses a function which is define by the model below"""
        if self.custom_set_function:
            key, value = self.custom_set_function(field_name, value)
        self.set_item(field_name, value)

    def set_geometry(self, geometry):
        """sets and converts the geometry"""
        if "the_geom" in self.geometry_fields.items:
            field_id = self.geometry_fields.translate["the_geom"]
            self.feature.SetGeomField(field_id, geometry)
        else:
            self.feature.SetGeometry(geometry)

        self.layer.SetFeature(self.feature)

    def set_item(self, field_name, value):
        """
        General routine for __setitem__,
        specific routines are added in the models themselves
        """

        # set the geometry
        if field_name in ["geometry", "the_geom"]:
            self.set_geometry(value)
            return

        self.feature.SetField(field_name, self.correct_type(field_name, value))
        self.layer.SetFeature(self.feature)

    def evaluate_types(self) -> dict:
        """
        Checks and changes field types are corresponding to 3Di types.
        In memory types can be slightley adjusted.
        returns a new dictionary.
        """

        return {k: self.correct_type(k, v) for k, v in self}

    def correct_type(self, field_name, value):
        """
        Loading in memory causes some type changes. Therefore we need to
        correct the types to get a good 3Di model.

        If the field is optional and is an 'id' field and is 0
        it should be None.

        Other types are compared with a standard types in constants.py

        Returns a correct value.
        """

        # This is what we should have.
        output_type = OGR_TYPES_INVERT[self.properties[field_name]["type"]]

        # First we check if the value we have should be None.
        # Integers
        is_optional = self.properties[field_name]["optional"]
        if is_optional and type(value) == int and value == 0:
            return None

        # floats # TODO this is pretty tricky since sometimes we want 0.0?
        if is_optional and type(value) == float and value == 0.0:
            return None

        # strings
        if is_optional and type(value) == str and value == "None":
            return None

        # Second we check if None is allowed.
        if value == None:
            if is_optional == True:  # none allowed
                return value
            else:
                raise ThreediValueTypeError(
                    "Value None is not allowed here",
                    self.table_name,
                    field_name,
                    value,
                )

        # Third we correct to a string, integer, float or bool.
        type_value = type(value)
        correct_type = output_type == type(value)

        if correct_type:
            return value

        if not correct_type:
            logger.debug(
                ThreediValueTypeWarning(
                    f"incorrect type, trying to correct {type_value} to {output_type}",
                    self.table_name,
                    field_name,
                    value,
                )
            )
            try:
                if output_type == bool:
                    value = int(value)
                value = output_type(value)
            except Exception:
                raise ThreediValueTypeError(
                    f"""Found incorrect type of {type_value} where  {output_type} is expected""",
                    self.table_name,
                    field_name,
                    value,
                )
            else:
                logger.debug(
                    ThreediValueTypeWarning(
                        f"{_HAPPY}, corrected input to {type(value)}: {value}",
                        self.table_name,
                        field_name,
                        value,
                    )
                )
                return value

    # lines = ['{']
    # for key, value in hoorn.global_setting.items.items():
    #     lines.append('{}:{}'.format(key, value))
    # lines.append(['}'])



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 13 09:15:34 2021

@author: chris.kerklaan

Example of sampling linestrings

"""

# imports
import os
from osgeo import ogr
import numpy as np
from threedi_edits import Vector, Raster, Progress, LineString, RasterGroup


def split(levees: Vector, distance=5):
    """splitting linestrings with a certain distance"""
    # first cut into pieces of 5 meter and sample some dem values
    split = levees.copy(shell=True)
    for levee in Progress(levees, "Splitting"):
        # create single levees due to multilinestrings
        for single_levee in levee.geometry.to_single():
            # split the levee in sections of 5 meter
            levee_lines = LineString(single_levee).split(distance)
            for levee_line in levee_lines:
                split.add(geometry=levee_line, items=levee.items)
    return split


def sample(levees: Vector, dem: Raster, distance=2, perpendicular_length=10):
    """
    Sample levees on a dem. Every distance a perpendicuar length is taken.
    The maxium on every perpendicular linestring is taken an averaged.

    params:
        levees: Vector
        dem: Raster or Rastergroup
        distance: distance between perpendicular lines
        perpendicular_lenhgth: Length of the perpendicular linestrings on
        which the sample is taken.

    """

    sampled = levees.copy(shell=True)
    sampled.add_field("height", float)
    for levee in Progress(levees, "Sampling"):
        lines = levee.geometry.perpendicular_lines(distance, perpendicular_length)
        heights = []
        for line in lines:
            line_data = dem.read(line)
            try:
                if len(line_data) > 0:
                    heights.append(np.nanmax(line_data))
            except Exception as e:
                print(e)
                pass
        mean = float(np.nanmean(heights))
        sampled.add(items=levee.items, height=mean, geometry=levee.geometry)

    return sampled


class LeveeAggregator:
    def __init__(
        self,
        levees: Vector,
        height_field: str = "height",
        mode: str = "min_max",
        threshold: float = 0.25,
        type_field: str = None,
        thresholds: dict = None,
        feature: bool = False,
        feature_dem: Raster = None,
        feature_split_distance: int = 10,
        feature_perpendicular_distance: int = 2,
        feature_perpendicular_length: int = 10,
    ):
        """
        Class to aggregate levees based on a field.
        Using 'type_field' it is possible to run the algorithm
        per unique field value. 'mode' sets the method in which the levees
        are aggregated.

        If the aggregator is run per feature, it has to split and sample the features
        on the go. So, the levee input does not have to be splitted and sampled which
        is the case if the aggregator is run differently.

        There are generally two ways in which  it is run:

           Without fields or worrying about features.
           ------
           aggregator = LeveeAggregator(levees, height_field='height', threshold=0.25)
           output = aggregator.run()
           output.write(output_path)

           With fields and not worrying about features.
           ------
           thresholds = {
               "kering-primair": 0.25,
               "kering-secundair": 0.25,
               "weg": 0.5,
               "spoor": 0.5,
           }
           aggregator = LeveeAggregator(levees, type_field='type', height_field='height', thresholds=thresholds)
           output = aggregator.run()
           output.write(output_path)

           With fields and using it per feature.
           ------
           thresholds = {
               "kering-primair": 0.25,
               "kering-secundair": 0.25,
               "weg": 0.5,
               "spoor": 0.5,
           }
           aggregator = LeveeAggregator(levees, type_field='type', height_field='height', thresholds=thresholds,
                                        feature=True,
                                        feature_dem=dem,
                                        feature_split_distance,
                                        feature_perpendicular_distance,
                                        feature_perpendicular_length)
           output = aggregator.run()
           output.write(output_path)




        params:
            levees: Vector with LineStrings, with or without height field
            height_field: Field in which the height is stored
            type_field: Field in which the vector gets separated.
            mode: Can either be 'min_max' or median.

            min_max:
            1. The adjacent levee height is not lower than the threshold minus the
            minium of the current heights.
            2. The adjecent levee height is not lower than the threshold plus the
            maximum of the current heights

            median:
            1. The difference between the median of the current heights
            and the next levee height is not larger than the threshold.

            feature: levees are converged per feature.
            feature_dem: Raster on which the levees are sampled,
            feature_split_distance: distance on which the levees are split,
            feature_perpendicular_distance: distance on wici perpedicular lines are created,
            feature_perpendicular_length: length of the perpendicular lines, we sample on this.

        """

        # making sure there are only np.nan's
        if not feature:
            data = levees.copy(shell=True)
            for levee in levees:
                items = levee.items
                if items[height_field] is None:
                    items[height_field] = np.nan
                data.add(items=items, geometry=levee.geometry)
        else:
            data = levees.copy()

        # separate the data if type_field is used
        if type_field:
            field_data = {}
            for value in list(set(data.table[type_field])):
                field_data[value] = data.filter(
                    return_vector=True, **{type_field: value}
                )
            self.field_data = field_data

        # adding to self to use in other functions
        self.data = data
        self.mode = mode
        self.threshold = threshold
        self.thresholds = thresholds
        self.feature = feature
        self.type_field = type_field
        self.dem = feature_dem
        self.feature_split_distance = (feature_split_distance,)
        self.feature_perpendicular_distance = feature_perpendicular_distance
        self.feature_perpendicular_length = feature_perpendicular_length

    def add_fields(self, output):
        output.add_field("m_count", int)
        output.add_field("m_aheight", float)
        output.add_field("m_max", float)
        output.add_field("m_min", float)
        output.add_field("height", float)
        return output

    def run(self):
        """run with standard settings"""
        if self.type_field:
            return self._run_fields(
                threshold=self.threshold, thresholds=self.thresholds
            )
        return self._run(self.data, self.threshold)

    def _run_fields(self, threshold=0.25, thresholds={}):
        field_output = self.data.copy(shell=True)
        field_output = self.add_fields(field_output)
        for field, data in self.field_data.items():
            if thresholds:
                threshold = thresholds[field]

            print("Running:", field, "Threshold:", threshold, "Input count", data.count)
            if self.feature:
                output = self._run_feature(
                    data,
                    self.dem,
                    threshold,
                    self.feature_split_distance,
                    self.feature_perpendicular_distance,
                    self.feature_perpendicular_length,
                )
            else:
                output = self._run(data, threshold)

            print("Output count:", output.count)
            for feature in output:
                field_output.add(feature)
        return field_output

    def _run(self, data, threshold=0.25):
        output = data.copy(shell=True)
        output = self.add_fields(output)

        data = data.copy()
        data.reset()

        levee = next(data)
        levee_geometry = levee.geometry
        levee_items = levee.items
        heights = [levee["height"]]

        data.delete(levee)

        while len(data) != 0:
            print(len(data), "levee_id:", levee.id)
            data.reset()
            levee_geometry, added_levees = add_adjacent_levees(
                heights, levee_geometry, data, threshold, self.mode
            )

            # add to output if nothing has to be added anymore
            if len(added_levees) == 0:

                output.add(
                    geometry=levee_geometry,
                    items=levee_items,
                    m_count=len(heights),
                    m_aheight=np.nanmean(heights),
                    m_max=np.nanmax(heights),
                    m_min=np.nanmin(heights),
                )

                levee = next(data)  # retrieves a new feature even if some are deleted
                levee_items = levee.items
                levee_geometry = levee.geometry
                heights = [levee["height"]]
                data.delete(levee)

            else:
                # delete from data and to heights
                for levee_id in added_levees:
                    added_levee = data[levee_id]
                    heights.append(added_levee["height"])
                    data.delete(added_levee)

        data.close()

        # finally
        output.add(
            geometry=levee_geometry,
            items=levee_items,
            m_count=len(heights),
            m_aheight=np.nanmean(heights),
            m_max=np.nanmax(heights),
            m_min=np.nanmin(heights),
        )

        return output

    def _run_feature(
        self,
        levees,
        dem,
        threshold,
        split_distance=10,
        perpendicular_distance=2,
        perpendicular_length=10,
    ):
        """the same thing as above but then per feature and splitting and sampling at the same time"""

        output = self.data.copy(shell=True)
        output = self.add_fields(output)

        # copy the original
        # get the first params and delete the first levee:
        #   levee_geometry
        #   levee_items
        #   heights

        # do it per feature
        for levee in levees:

            print(levee.id)
            # split the levee into sections
            print("Splitting")
            split_lines = []
            for single_levee in levee.geometry.to_single():
                split_lines.extend(LineString(single_levee).split(split_distance))

            # sample the data on the dem
            print("Sampling")
            sampled_lines = []
            for line in Progress(split_lines, "Sampling"):

                heights = []
                lines = line.perpendicular_lines(
                    perpendicular_distance, perpendicular_length
                )
                for line in lines:
                    line_data = dem.read(line)
                    try:
                        if len(line_data) > 0:
                            heights.append(np.nanmax(line_data))
                    except Exception as e:
                        print(e)
                        pass

                sampled_lines.append(float(np.nanmean(heights)))

            # merge into a vector
            print("Merging")
            vector = output.copy(shell=True)
            for geometry, height in zip(split_lines, sampled_lines):
                vector.add(items=levee.items, geometry=geometry, height=height)

            # run the convergence
            print("Converging")
            feature_output = self._run(vector, threshold)
            for feature in feature_output:
                output.add(feature)

        return output


def add_adjacent_levees(
    heights: list,
    levee_geometry: ogr.Geometry,
    levee_vector: Vector,
    threshold=0.25,
    mode="min_max",
):
    """
    adds adjacent levees if based on a list of previously generated
    heights and the height of the adjecent levee.
    The adjacent levee and its height are added in the following cases:

        min_max:
        1. The adjacent levee height is not lower than the threshold minus the
        minium of the current heights.
        2. The adjecent levee height is not lower than the threshold plus the
        maximum of the current heights

        median:
        1. The difference between the median of the current heights
        and the next levee height is not larger than the threshold.

    other params:
        heights: list of heights
        levee_geometry: geometry of merged levees
        levee_vector: levee vector


    """

    output_geometry = levee_geometry.copy()
    added_levees = []
    for levee in levee_vector.spatial_filter(levee_geometry.buffer(0.0001)):
        next_levee_height = levee["height"]

        is_nan = np.isnan(next_levee_height)
        if not is_nan:
            if mode == "min_max":
                minimum_difference = abs(next_levee_height - min(heights)) < threshold
                maximum_difference = abs(next_levee_height - max(heights)) < threshold
                add = minimum_difference and maximum_difference
            elif mode == "median":
                median = np.percentile(heights, 50)
                add = abs(median - next_levee_height) < threshold
        else:
            add = False

        if add or is_nan:
            added_levees.append(levee.id)
            output_geometry = output_geometry.union(levee.geometry)
            output_geometry = output_geometry.dissolve()

    return output_geometry, added_levees


if __name__ == "__main__":
    # from here the code is run
    os.chdir(
        r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing"
    )

    # load the levees
    levees = Vector("levees/all_levees.gpkg")

    # load the dem as a rastergroup, because they are in tiles
    dem = RasterGroup([Raster(f"data/dtm/{i}") for i in os.listdir("data/dtm")])

    # step 1 - split the levees
    split_levees = split(levees, distance=10)
    split_levees.write("levees/split_10m.gpkg")

    # step 2 - sample the levees on the dem
    sampled_levees = sample(split_levees, dem, distance=2, perpendicular_length=10)
    sampled_levees.write("levees/sampled_10m.gpkg")

    # step 3 - run the aggregator
    levees.add_field("height", float)
    aggregator = LeveeAggregator(levees, height_field="height", mode="min_max")
    levees_aggregated = aggregator.run(aggregator.data, 0.25)
    levees_aggregated.write("levees/aggregated.gpkg")

    # or everything at the same time
    thresholds = {
        "kering-primair": 0.25,
        "kering-secundair": 0.25,
        "weg": 0.5,
        "spoor": 0.5,
    }

    aggregator = LeveeAggregator(
        levees,
        type_field="type",
        mode="median",
        thresholds=thresholds,
        feature=True,
        feature_dem=dem,
    )
    levees_aggregated = aggregator.run()
    levees_aggregated.write("levees/aggregated_per_feature2.gpkg")



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 14 14:20:45 2021

@author: chris.kerklaan

Example shows how to merge multiple vector files

"""
import os
from threedi_edits import Vector, Progress

os.chdir(r"C:\Users\chris.kerklaan\Documents\Projecten\basis kaarten hhnk\processing")

# define paths
dikes_path = "data/waterkering_werkbestand.shp"
roads_path = "data/wegen.gpkg"
rail_path = "data/spoorwegen.gpkg"

# load data
dikes = Vector(dikes_path)
roads = Vector(roads_path)
rails = Vector(rail_path)

# create output -- help
help(Vector.from_scratch)

# create a vector with the name levees, a linstring and projection of the Netherlands
output = Vector.from_scratch("levees", 2, 28992)

# add a field to track which type of data is added
output.add_field("type", str)

# add dike keep geometry and 'code', 'category' and 'name'
# first we add fields to the output
output.add_field("waterkering_code", str)
output.add_field("waterkering_naam", str)
output.add_field("waterkering_categorie", str)
output.add_field("weg_fysiekvoorkomen", str)
output.add_field("weg_typeweg", str)
output.add_field("weg_id", int)
output.add_field("weg_naam", str)

output.add_field("spoor_fuuid", str)
output.add_field("spoor_geocode_naam", str)


# show progress as well
for dike in Progress(dikes, "Adding dikes"):
    # extract data
    dike_code = dike["CODE"]
    dike_name = dike["NAAM"]
    dike_category = dike["CATEGORIE"]

    # transform to new names
    if dike_category == 1:
        type = "kering-primair"
    else:
        type = "kering-secundair"
    items = {
        "waterkering_code": dike_code,
        "waterkering_naam": dike_name,
        "waterkering_categorie": dike_category,
        "type": type,
    }
    # 3D geom
    dike_geometry = dike.geometry.fix()
    # add to output
    output.add(items=items, geometry=dike_geometry)


# add roads keep geometry and 'fysiekvoorkomen', 'typeweg', 'id', 'name'
# first we add fields to the output

# show progress as well
for road in Progress(roads, "Adding roads"):
    # extract data
    road_id = road["id"]
    road_name = road["Naam"]
    road_fysiekvoorkomen = road["fysiekvoorkomen"]
    road_type = road["typeweg"]

    # transform to new names
    items = {
        "weg_fysiekvoorkomen": road_fysiekvoorkomen,
        "weg_typeweg": road_type,
        "weg_id": road_id,
        "weg_naam": road_name,
        "type": "weg",
    }

    # add to output
    output.add(items=items, geometry=road.geometry)


# show progress as well
for rail in Progress(rails, "Adding rails"):

    # extract data
    rail_name = rail["geocode_naam"]
    rail_id = rail["fuuid"]

    # transform to new names
    items = {"spoor_fuuid": rail_id, "spoor_geocode_naam": rail_name, "type": "spoor"}

    # add to output
    output.add(items=items, geometry=rail.geometry)

# write as a geopackeg to keep the names
output.write("levees/all_levees.gpkg")



# -*- coding: utf-8 -*-
"""
Created on Fri Nov 26 10:30:14 2021

@author: chris.kerklaan
"""

import sys
import types
import datetime


class Progress:
    def __init__(
        self,
        iterator=None,
        message="",
        quiet=False,
        enum=False,
        total=None,
        yield_progress=False,
        size=30,
        file=sys.stdout,
        start_offset=None,
    ):
        self.iterator = iterator
        self.message = message
        self.size = size
        self.file = file
        self.time = datetime.datetime.now()
        self.enumerate = enum
        self.yield_progress = yield_progress

        if start_offset:
            self.i = start_offset
        else:
            self.i = 0

        if isinstance(iterator, types.GeneratorType) and total == None:
            raise ValueError("Type is generator, please fill in total")

        if total:
            self.total = total
        else:
            self.total = len(iterator)

        self.quiet = quiet

    def __iter__(self):
        if self.enumerate:
            for i, x in enumerate(self.iterator):
                self.show()
                if self.yield_progress:
                    yield self.progress, i, x
                else:
                    yield i, x
        else:
            for i in self.iterator:
                self.show()

                if self.yield_progress:
                    yield self.progress, i
                else:
                    yield i

    def __next__(self):
        yield self.iterator.__next__()

    @property
    def progress(self):
        """progress out of a 100"""
        return self.progress_factor * 100

    @property
    def progress_factor(self):
        return self.i / self.total

    def gdal(self, complete, message=None, cb_data=None):
        """Emit progress report in numbers for 10% intervals and dots for 3%"""
        self.i = int(self.total * complete)
        self.show(False)

    def show(self, add=True, custom_add=0):
        """shows the bar, count can be increased if needed
        params:
            add: automatically adds 1 to the progress
            custom_add: can add more or less than 1 to progress
        """
        if add:
            if custom_add != 0:
                self.i += custom_add
            else:
                self.i += 1

        if self.quiet:
            return

        x = int(self.size * self.progress_factor)

        print(
            f"\r{self.message}: [{'.'*x}{' '*(self.size-x)}] {self.i}/{self.total}",
            end="\r",
            flush=True,
        )

        if self.i >= self.total:
            self.show_end()

    def show_end(self):
        end_time = datetime.datetime.now() - self.time
        # next line
        print("\r")
        # then some message
        print(
            f"\r{self.message}: Completed in {end_time} n = {self.total}",
            end="\r\n",
            flush=True,
        )

    def update(self):
        self.show()
        if self.yield_progress:
            return self.progress



# -*- coding: utf-8 -*-
"""
Created on Tue Feb 23 14:14:38 2021

@author: chris.kerklaan
"""
# First-party imports
import logging

# Third-party imports
from osgeo import ogr

# optional imports
import importlib

HAS_SCIPY = importlib.util.find_spec("scipy") is not None
if HAS_SCIPY:
    from scipy import spatial

# Local imports
from .geometry import dissolve, Geometry
from .linestring import LineString, MultiLineString

# GLOBALS
logger = logging.getLogger(__name__)

POLYGON_COVERAGE = [
    ogr.wkbPolygon,
    ogr.wkbPolygon25D,
    ogr.wkbPolygonM,
    ogr.wkbPolygonZM,
    ogr.wkbCurvePolygon,
    ogr.wkbCurvePolygonM,
    ogr.wkbCurvePolygonZ,
    ogr.wkbCurvePolygonZM,
]

MULTIPOLYGON_COVERAGE = [
    ogr.wkbMultiPolygon,
    ogr.wkbMultiPolygon25D,
    ogr.wkbMultiPolygonM,
    ogr.wkbMultiPolygonZM,
    ogr.wkbMultiSurface,
    ogr.wkbMultiSurface,
    ogr.wkbMultiSurfaceM,
    ogr.wkbMultiSurfaceZ,
    ogr.wkbMultiSurfaceZM,
]


class Polygon(Geometry):
    ogr_coverage = POLYGON_COVERAGE

    def __init__(self, geometry: ogr.wkbPolygon = None, epsg=None):
        super().__init__(geometry)
        self.check_type(Polygon.ogr_coverage)
        if epsg:
            self.spatial_reference = epsg

    @classmethod
    def from_points(cls, points, flatten=True, close=True, epsg=None):
        """takes list of tuple points and creates an ogr polygon"""

        output_geom = ogr.Geometry(ogr.wkbPolygon)

        ring = ogr.Geometry(ogr.wkbLinearRing)
        for point in points:
            ring.AddPoint(*point)

        output_geom.AddGeometry(ring)

        if close:
            output_geom.CloseRings()

        if flatten:
            output_geom.FlattenTo2D()

        if not output_geom.IsValid():
            logger.warning(
                """
                            Is it a self-intersection polygon?
                            Are the points in the form a ring? E.g., 
                            left-upper, left-lower, right-lower, right-upper"""
            )

        return cls(output_geom, epsg)

    @classmethod
    def from_wkt(cls, wkt, epsg=None):
        """takes tuple points and creates an ogr point"""
        return cls(ogr.CreateGeometryFromWkt(wkt), epsg)

    @property
    def narrowity(self):
        """
        returns the area of the geometry divided by the length/perimeter.

        This is generally used as a pre-filter to derive long and narrow polygons.
        Interpretation depends on the use

        """

        return self.boundary.length / self.area

    @property
    def narrow(self):
        """
        returns True if self.narrowity is smaller than 0.35
        """
        return self.narrowity > 0.35

    def centerline(self, resolution=0.5, intersection=True):
        """
        returns the center of this polygon
        params:
            resolution: resolution of the voronoi boundary points
            intersection: cuts the multiline at intersections

        A centerline is returned by creating points on the boundary of the line.
        The points are used to create vonoroi polygons.
        Filtering on the ridges inside the polygon returns the centerline

        """
        return MultiLineString(centerline(self, resolution, intersection))

    def width(self, samples: int):
        """
        returns a list of possible widths, made along the centerline of the polygon
        params:
            samples: amount of samples taken along the centerline

        The direction of a polygon is hard to describe. We describe it as
        the centerline of the polygon. Perpendicular lines are taken
        and clipped to the geometry to sample some widths of the polygon.
        Which is returned in this function

        It is a very costly operation, so use with care.

        """
        return width(self, samples)


class MultiPolygon(Geometry):
    ogr_coverage = MULTIPOLYGON_COVERAGE

    def __init__(
        self, geometry: ogr.wkbMultiPolygon = None, points: list = None, epsg=None
    ):

        if points:
            geometry = self.create_multipolygon(points)

        super().__init__(geometry)

        if epsg:
            self.spatial_reference = epsg

        # check types
        if self.type == ogr.wkbPolygon:
            raise TypeError(
                """geometry is multi while vector is single
                                use vector.to_single()
                            """
            )
        self.check_type(MultiPolygon.ogr_coverage)

    @classmethod
    def from_points(cls, points, flatten=True, close=True, epsg=None):
        output_geom = ogr.Geometry(ogr.wkbMultiPolygon)

        for point in points:
            ring = ogr.Geometry(ogr.wkbLinearRing)
            for p in point:
                ring.AddPoint(p)

            poly = ogr.Geometry(ogr.wkbPolygon)
            poly.AddGeometry(ring)
            output_geom.AddGeometry(poly)

        if close:
            output_geom.CloseRings()

        if flatten:
            output_geom.FlattenTo2D()

        if not output_geom.IsValid():
            logger.warning(
                """
                            Is it a self-intersection polygon?
                            Are the points in the form a ring? E.g., 
                            left-upper, left-lower, right-lower, right-upper"""
            )
        return cls(output_geom, epsg)

    @classmethod
    def from_wkt(cls, wkt, epsg=None):
        """takes tuple points and creates an ogr point"""
        return cls(ogr.CreateGeometryFromWkt(wkt), epsg)


def centerline(geometry: Polygon, resolution: int, intersection: bool, clean=False):
    if not HAS_SCIPY:
        raise ImportError("Please install scipy before using this function")

    boundary = geometry.boundary
    points = LineString(boundary).points_on_line(resolution)

    vor = spatial.Voronoi([p.point for p in points])
    vertices = vor.vertices
    ridges = vor.ridge_vertices

    multi = ogr.Geometry(5)
    for ridge in ridges:
        if "-1" not in ridge:  # is infinite
            start_point = (vertices[ridge[0]][0], vertices[ridge[0]][1])
            end_point = (vertices[ridge[1]][0], vertices[ridge[1]][1])
            line = LineString.from_points([start_point, end_point])

            if line.Within(geometry):
                multi.AddGeometry(line)

    dissolved = dissolve(multi)
    if clean:
        return MultiLineString(dissolved).clean(intersection)
    return MultiLineString(dissolved)


def width(geometry: Polygon, samples: int = 10):
    centerlines = geometry.centerline(0.5, intersection=True)

    # keep the longest line
    for i, centerline in enumerate(centerlines):
        if i == 0:
            line = centerline
        else:
            if centerline.Length() > line.Length():
                line = centerline

    line = LineString(line)
    perpendicular_lines = line.perpendicular_lines(
        line.length / (samples + 1), line.length, start=False, end=False
    )

    widths = []
    for line in perpendicular_lines:
        intersect = line.clip(geometry)
        if intersect:
            widths.append(intersect[0].length)
    return widths



# -*- coding: utf-8 -*-
"""
Created on Thu Feb 25 11:38:51 2021

@author: chris.kerklaan


Script for automated testing of the raster group object


Currently supported functions
    1. Check alignment
        1. Data/nodata value
        2. geotransform
        3. rows, columns
        
    

"""
# First-party imports
import pathlib

# Third-party imports
import numpy as np

# Local imports
from threedi_edits.gis.raster import Raster
from threedi_edits.gis.rastergroup import RasterGroup

# Globals
DATA = pathlib.Path(__file__).parent / "data" / "raster" / "groups"


RASTERS = []
for raster_path in DATA.glob("*.tif"):
    print(raster_path)
    raster = Raster(raster_path)
    RASTERS.append(raster.copy(shell=False, memory=False))


def test_check_alignment_geotransform():
    """Tests is alignment checking notices geotransform changes"""
    test_rasters = [raster.copy() for raster in RASTERS]
    test_rasters[1].geotransform = (96798.0, 0.5, 0.0, 417515.0, 0.0, -0.5)
    rastergroup = RasterGroup(test_rasters)

    output = rastergroup.check_alignment()
    assert len(output["errors"]) > 0


def test_check_alignment_rows_columns():
    """Tests is alignment checking notices rows/columns changes"""
    test_rasters = [raster.copy() for raster in RASTERS]
    new = Raster.from_scratch(
        "new",
        test_rasters[1].geotransform,
        1,
        1,
    )
    test_rasters.append(new)
    rastergroup = RasterGroup(test_rasters)

    output = rastergroup.check_alignment()
    assert len(output["errors"]) > 0


def test_check_alignment_data_nodata():
    """Tests is alignment checking notices data/nodata changes"""
    test_rasters = [raster.copy() for raster in RASTERS]
    array = test_rasters[2].array

    array[array > 0] = np.nan
    test_rasters[2].array = array
    test_rasters[2].name = "Test raster"
    rastergroup = RasterGroup(test_rasters)

    output = rastergroup.check_alignment()
    errors_present = len(output["errors"]) > 0
    assert errors_present

    if errors_present:
        assert "counts" in output["errors"][0]


def test_tiling():
    """tests if tiling works"""
    test_rasters = [raster.copy() for raster in RASTERS]
    group = RasterGroup(test_rasters)

    for tiles in group.tiles():
        assert tiles[0].shape == (1, 1032)
        break



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 30 13:52:55 2019

@author: chris.kerklaan
"""
from threedi_edits.threedi.constants.v0209.templates import *  # noqa
from threedi_edits.threedi.constants.v0209.constants import *  # noqa



# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 18:20:00 2020

@author: chris.kerklaan

Raster wrapper used for gis programming
This raster wrapper has as peculiarity that for every function the raster is 
copied/loaded into memory. Hence it always has to be written. I.e.,
it cannot be edited in place.

TODO:
    1. Buffer toevoegen aan een raster --> Misschien iets voor rastergroup ThreeDi
    4. Optimalizeren van blocks
    5. Testen voor dit hele script, testen moeten ook kijken naar geheugen issuses
    
Notes:
    Memory handling:
        1. Numpy calculations are done in tiles as much as possible - speed/memory efficiency
        2. This is also why rasters are edited in place and not returned as new rasters.
        
    Bottleneck:
        1. Rasters are loaded into memory if needed, before the use of functions 
        2. Due to the fact that the rasters are loaded into memory, performance depends on your system
        
    Functions:
        1. Each function handles the object in tiles.
        2. After each function the old file is released and the file is referenced.
        3. Functions outside of the class or either extremely complicated or
           also use multiple external functions
    
    Iterating:
        1. One can iterate over this raster using for tile in raster:
           Blocksizes are optimzed and arrays must be closed after using
           Recommeded code:
               output = raster.copy(shell=True)
               for tile in raster:
                   array = tile
                   # do something with array
                   output.array = array, *tile.location
                   

"""
# First-party import
import logging
import os
import pathlib
import importlib
import tempfile
import numpy as np

# Third-party imports
from osgeo import ogr, gdal, gdal_array

# Third-party imports optional

# osgeo - base package
HAS_SCIPY = importlib.util.find_spec("scipy") is not None
if HAS_SCIPY:
    import scipy

# Local imports
from .vector import Vector
from .geotransform import GeoTransform
from .utils import Layer, Dataset
from .geometry import POLYGON
from .spatial_reference import SpatialReference
from threedi_edits.utils.progress import Progress
from threedi_edits.utils.project import Classes, Functions

# structure
classes = Classes(__name__, local_only=True)
functions = Functions(__name__)


# GLOBALS
# Drivers
DRIVER_OGR_SHP = ogr.GetDriverByName("Esri Shapefile")
DRIVER_GDAL_GTIFF = gdal.GetDriverByName("GTiff")
DRIVER_GDAL_MEM = gdal.GetDriverByName("MEM")

# Options
WRITE_OPTIONS = ["COMPRESS=DEFLATE", "NUM_THREADS=4", "ZLEVEL=9", "TILED=YES"]

# Gdal/ogr band/field mapping
TYPE_MAPPING = {
    gdal.GDT_Byte: ogr.OFTInteger,
    gdal.GDT_UInt16: ogr.OFTInteger,
    gdal.GDT_Int16: ogr.OFTInteger,
    gdal.GDT_UInt32: ogr.OFTInteger,
    gdal.GDT_Int32: ogr.OFTInteger,
    gdal.GDT_Float32: ogr.OFTReal,
    gdal.GDT_Float64: ogr.OFTReal,
    gdal.GDT_CInt16: ogr.OFTInteger,
    gdal.GDT_CInt32: ogr.OFTInteger,
    gdal.GDT_CFloat32: ogr.OFTReal,
    gdal.GDT_CFloat64: ogr.OFTReal,
}
TYPE_MAPPING_INPUT = {
    float: gdal.GDT_Float32,
    int: gdal.GDT_Int32,
}
# memory counter
_loc_ras = 0

# logging
logger = logging.getLogger(__name__)


class Raster:

    instance = "gis.raster.Raster"

    def __init__(
        self,
        raster_input,
        name=None,
        optimize_blocksize=True,
        write=0,
        **kwargs,
    ):
        """

        Raster -- info

        Raster object for general raster GIS processing and reading
        Raster has functions such as:
            - read(geometry)
            - clip
            - merge
            - smooth
            - resample
            - polygonizing
            - push_vector
            - align
            - fill_nodata
            - reproject

        It also uses 'raster looping':
            Iterating:
                1. One can iterate over this raster using for tile in raster:
                    Blocksizes are optimzed and arrays must be closed after using
                    Recommended code:
                        output = raster.copy(shell=True)
                        for tile in raster:
                            array = tile.array
                            # do something with array
                            output.array = array, *tile.location
                        ouput.write(path)

            Inflated iterating with a buffer of 1:
                for tile, original_extent in raster.inflated_tiles(1, True):
                    tile.array
                    # do something with array
                    tile.clip_bbox(*original_extent)

        Use 'help()' for more info on the functions

        params:
            raster_input: str, pathlib.path or gdal.ds
            name: if None, extracted from path
            blocksize: blocksize for raster looping
            optimize_blocksize: if False, uses a standard blocksize
            **kwargs: geotransform, spatial_reference and nodata

        options:
            self.return_raster: Boolean Returns a Raster object at every function
                                instead of inplace changes
            self.nan_array: Boolean Sets all nodata to nan, usefull for raster calculations


        """
        if isinstance(raster_input, pathlib.PurePath):
            ds, name = self.from_path(raster_input, name, write)
        elif type(raster_input) is str:
            ds, name = self.from_path(raster_input, name, write)
        else:
            ds = raster_input

        if not name:
            name = "No name"

        self.ds = ds
        self.name = name

        # options
        self.return_raster = False
        self.nan_array = True
        self.optimize_blocksize = optimize_blocksize

        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def from_array(cls, array, name=None, **kwargs):
        """import from an array raster array"""
        cls.type = "array"
        if len(array.shape) != 2:
            raise ValueError("limited to 2d arrays, check array.shape")
        return cls(gdal_array.OpenArray(array), name, **kwargs)

    @classmethod
    def from_scratch(
        cls,
        name,
        geotransform,
        columns,
        rows,
        nodata_value=-9999,
        data_type=gdal.GDT_Float32,
        epsg=28992,
        memory=True,
    ):
        out, _ = create_source(columns, rows, data_type, memory)

        return cls(
            out,
            name=name,
            geotransform=geotransform,
            spatial_reference=epsg,
            nodata_value=nodata_value,
        )

    def from_path(self, path, name, write=0):
        if not os.path.exists(path):
            raise ValueError(f"Path does not exist: {path}")

        self.type = "path"
        if not name:
            name = pathlib.Path(path).stem
        return gdal.Open(str(path), write), name

    def __iter__(self):
        generator = tile_generator(
            self.ds, self.optimize_blocksize, self.blocksize[0], self.blocksize[1]
        )
        for block, location in generator:
            raster = Raster(block, name=f"{location} - {self.name}")
            raster.location = location
            yield raster

    def tiles_inflated(self, buffer_size=1, original_extent=False):
        """Generator for inflated blocks projection"""
        for or_ext, tile in zip(self.tile_extents, self):
            buffered = tile.extent_vector.buffer(buffer_size)

            clipped = self.clip_bbox(*buffered.extent, name=tile.name)
            if original_extent:
                yield clipped, or_ext
            else:
                yield clipped

    def __len__(self):
        return len(self.tiles)

    def __setitem__(self, key, value):
        x, y = key
        return self.band.WriteArray(value, x, y)

    def __getitem__(self, key):
        x, y = key
        return self.band.ReadAsArray(x, y, self.w, self.h)

    def __str__(self):
        return self.name

    def __repr__(self):
        settings = self.settings
        settings["size"] = self.size
        settings["resolution"] = self.resolution
        return f"({self.instance}) {self.name}: {self.settings}"

    @property
    def array(self):
        array = self.band.ReadAsArray()
        if self.nan_array:
            return self.nan(array)
        return array

    @array.setter
    def array(self, value: (np.ndarray, 0, 0)):
        """Takes '(numpy array, x, y)' to set the array"""
        if type(value) != tuple:
            x = 0
            y = 0
            array = value
        else:
            array, x, y = value

        self.band.WriteArray(array, x, y)
        self.band.FlushCache()

    @property
    def name(self):
        if not hasattr(self, "_name"):
            return "no-name"
        else:
            return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def geotransform(self):
        return GeoTransform(self.ds.GetGeoTransform())

    @geotransform.setter
    def geotransform(self, value):
        self.ds.SetGeoTransform(value)

    @property
    def nodata_value(self):
        return self.band.GetNoDataValue()

    @nodata_value.setter
    def nodata_value(self, value):
        self.band.SetNoDataValue(value)

    @property
    def spatial_reference(self):
        return SpatialReference(wkt=self.ds.GetProjection())

    @spatial_reference.setter
    def spatial_reference(self, sr_input):
        """input is either epsg or sr object"""
        if type(sr_input) == int:
            self.ds.SetProjection(SpatialReference.from_epsg(sr_input).wkt)
        else:
            self.ds.SetProjection(sr_input.wkt)

    @property
    def metadata(self):
        return self.ds.GetMetadata()

    @metadata.setter
    def metadata(self, metadata: dict):
        self.ds.SetMetadata(metadata)

    @property
    def band(self):
        return self.ds.GetRasterBand(1)

    @property
    def epsg(self):
        return self.spatial_reference.epsg

    @property
    def mask(self):
        return ~np.isnan(self.array)

    @property
    def data_type(self):
        return self.band.DataType

    @property
    def data_type_name(self):
        return gdal.GetDataTypeName(self.data_type)

    @property
    def np_data_type(self):
        return gdal_array.GDALTypeCodeToNumericTypeCode(self.data_type)

    @property
    def columns(self):
        return self.band.XSize

    @property
    def rows(self):
        return self.band.YSize

    @property
    def pixels(self):
        return self.columns * self.rows

    @property
    def resolution(self):
        return {"width": self.geotransform[1], "height": self.geotransform[5]}

    @property
    def extent(self):
        return self.get_extent()

    @property
    def extent_geometry(self):
        return ogr.CreateGeometryFromWkt(
            POLYGON.format(x1=self.minx, x2=self.maxx, y1=self.miny, y2=self.maxy)
        )

    @property
    def extent_vector(self):
        v = Vector.from_scratch("tile", ogr.wkbPolygon, self.epsg)
        v.add(geometry=self.extent_geometry)
        return v

    @property
    def minx(self):
        return self.extent[0]

    @property
    def maxx(self):
        return self.extent[1]

    @property
    def miny(self):
        return self.extent[2]

    @property
    def maxy(self):
        return self.extent[3]

    @property
    def width(self):
        return self.maxx - self.minx

    @property
    def height(self):
        return self.maxy - self.miny

    @property
    def size(self):
        return self.rows * self.columns

    @property
    def shape(self):
        return (self.rows, self.columns)

    @property
    def blocksize(self):
        if not hasattr(self, "_blocksize"):
            self._blocksize = self.band.GetBlockSize()

        return self._blocksize

    @blocksize.setter
    def blocksize(self, value=(256, 256)):
        self._blocksize = list(value)
        self.optimize_blocksize = False

    @property
    def optimized_blocksize(self):
        return optimize_blocksize(self.ds)

    @property
    def tiles(self):
        if not hasattr(self, "_tiles"):
            self._tileopt = self.optimize_blocksize
            self._tiles = tile_list(self.ds, self._tileopt, *self.blocksize)

        if self.optimize_blocksize != self._tileopt:
            self._tileopt = self.optimize_blocksize
            self._tiles = tile_list(self.ds, self._tileopt, *self.blocksize)

        return self._tiles

    @property
    def tile_extents(self):
        return [t.extent_vector.extent for t in self]

    @property
    def settings(self):
        return {
            "spatial_reference": self.spatial_reference,
            "geotransform": self.geotransform,
            "nodata_value": self.nodata_value,
        }

    def load_to_memory(self):
        """loads this datasource to memory, usefull for boosting read array speed"""
        self.change_source(DRIVER_GDAL_MEM.CreateCopy("", self.ds), name=self.name)

    def close(self):
        self.ds.FlushCache()
        del self.ds

    def change_source(self, in_ds, name=None):
        """closes old source and references this object to a new souce"""
        self.ds = None
        self.ds = in_ds
        if name == None:
            self.name = "No name"
        else:
            self.name = name

    def copy(self, shell=False, data_type=None, memory=True):
        if not shell:
            path = loc_dir(mem=memory, add=True)
            if memory:
                driver = DRIVER_GDAL_MEM
            else:
                driver = DRIVER_GDAL_GTIFF

            return Raster(
                driver.CreateCopy(path, self.ds), memory=memory, **self.settings
            )

        if not data_type:
            data_type = self.data_type

        return Raster.from_scratch(
            self.name,
            self.geotransform,
            self.columns,
            self.rows,
            self.nodata_value,
            data_type,
            self.epsg,
            memory,
        )

    def replace_nodata(self, nodata_value=-9999, quiet=False):
        """replaces the in-array nodata values in a tiled form"""
        target = self.copy(shell=True)
        for tile in Progress(self, f"Replace nodata {self.nodata_value}", quiet):
            array = tile.array
            array[np.isnan(array)] = nodata_value
            target.array = array, *tile.location
            tile.close()

        target.nodata_value = nodata_value
        self.change_source(target.ds, name=self.name)

    def change_data_type(self, data_type=float):
        """changes the data type of raster using float, int or gdal dt"""
        if data_type in [float, int]:
            data_type = TYPE_MAPPING_INPUT[data_type]
        out, path = create_source(self.columns, self.rows, data_type)
        target = Raster(out, name=self.name, **self.settings)
        target.array = self.array
        self.change_source(target.ds, name=self.name)

    def nan(self, array):
        """returns an np array which the nodata values are set to np.nan"""
        if np.dtype(array.dtype).char in np.typecodes["AllInteger"]:
            array = array.astype(float)
        array[array == self.nodata_value] = np.nan
        return array

    def resize_band(self, width, height):
        self.band.WriteArray(np.full([height, width], np.nan))

    def get_extent(self):
        """returns the extent of the geotransform"""

        transform = self.geotransform
        maxx = transform[0] + self.columns * transform[1] + self.rows * transform[2]

        miny = transform[3] + self.columns * transform[4] + self.rows * transform[5]

        minx = transform[0]
        maxy = transform[3]
        return minx, maxx, miny, maxy

    def clip(self, vector=None, geometry=None, quiet=False):
        """
        Clips the raster based on an geometry.
        params:
            vector: Clips the raster on a vector.
            geometry: Clips the raster on a geometry.
            quiet: Verbosity.
        return:
            A clipped raster.
        """
        if geometry:
            vector = Vector.from_scratch("", geometry.type, geometry.epsg)
            vector.add(geometry=geometry)

        with tempfile.TemporaryDirectory() as tempdir:

            logger.debug("Clip using the vector path")

            vector.write(tempdir + "/vector.gpkg")
            progress = Progress(total=100, message="Clip", quiet=quiet)

            options = {
                "cutlineDSName": tempdir + "/vector.gpkg",
                "cutlineLayer": vector.layer_name,
                "cropToCutline": True,
                "multithread": True,
                "creationOptions": WRITE_OPTIONS,
                "callback": progress.gdal,
            }

            ds = gdal.Warp(loc_dir(mem=True), self.ds, **options)

        return Raster(ds, name=self.name)

    def clip_bbox(self, minx, maxx, miny, maxy, name=None):
        """clips on a raster on a bbox"""
        return Raster(
            gdal.Translate(
                loc_dir(mem=True, add=True), self.ds, projWin=(minx, maxy, maxx, miny)
            ),
            name=name,
        )

    def reproject(self, epsg, quiet=True):
        progress = Progress(total=100, message="Reproject", quiet=quiet)

        options = {
            "dstSRS": f"EPSG:{epsg}",
            "creationOptions": WRITE_OPTIONS,
            "callback": progress.gdal,
        }

        ds = gdal.Warp(loc_dir(mem=True), self.ds, **options)

        if self.return_raster:
            return Raster(ds)
        else:
            self.change_source(ds, name="reprojected")

    def resample(self, x_res, y_res, algorithm=gdal.GRA_NearestNeighbour, quiet=True):

        progress = Progress(total=100, message="Resample", quiet=quiet)
        options = {
            "xRes": x_res,
            "yRes": y_res,
            "resampleAlg": algorithm,
            "creationOptions": WRITE_OPTIONS,
            "callback": progress.gdal,
        }

        ds = gdal.Translate(loc_dir(mem=True), self.ds, **options)

        if self.return_raster:
            return Raster(ds)
        else:
            self.change_source(ds, name="resampled")

    def smooth(
        self,
        method="gaussian",
        arguments={"sigma": 1, "mode": "nearest"},
        quiet=False,
    ):
        """
        Smooths a raster based on two methods, returns a raster object
        params:
            method: 'gaussian' or 'perona_malik'
            arguments:
                gaussian:
                    - sigma
                    - mode
                    see https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.gaussian_filter.html
                perona_malik:
                    - iter
                    - K

        """

        if not HAS_SCIPY:
            raise ImportError("Please install scipy to use this option")

        target = self.copy(shell=True)
        if method == "gaussian":
            if not quiet:
                logger.debug(f"Applying gaussian filter with arguments: {arguments}")

            output_array = scipy.ndimage.gaussian_filter(
                self.array, output=self.np_data_type, **arguments
            )
        elif method == "perona_malik":
            output_array = perona_malik_filter(self.array, **arguments)

        target.array = output_array

        return target

    def push_vector(
        self,
        vector: Vector,
        value=None,
        field=None,
        set_to_nodata=False,
        quiet=False,
    ):
        """
        Pushes a vector this raster

        Params:
            Vector is a vector object
            Value (optional): pushes this value onto the raster
            Field (optional): pushes this field onto the raster
            set_to_nodata (optional): pushes nodata onto the raster

        """

        if self.epsg != vector.epsg:
            vector = vector.reproject(self.epsg)

        if value == self.nodata_value:
            set_to_nodata = True

        target = self.copy(shell=True)
        nodata = self.nodata_value

        # rasterize vector vector array
        rasterized_vector = Raster(
            vector.rasterize(
                rows=self.rows,
                columns=self.columns,
                geotransform=self.geotransform,
                nodata=nodata,
                field=field,
                return_ds=True,
            )
        )

        # make sure that it loops over the same blocks
        rasterized_vector.blocksize = self.blocksize
        self.optimize_blocksize = False

        for raster_tile, vector_tile in zip(self, rasterized_vector):
            array = raster_tile.array
            vector_array = vector_tile.array

            # if no value and field are filled, it is set to nodata?
            if set_to_nodata:
                array[vector_array == 1] = nodata
                target.array = (array, *raster_tile.location)

            elif value:
                array[~np.isnan(vector_array)] = value
                target.array = (array, *raster_tile.location)

            elif field:
                mask = ~np.isnan(vector_array)
                array[mask] = vector_array[mask]
                target.array = (array, *raster_tile.location)

            del array
            del vector_array

        return target

    def push_raster(self, raster, value=None):
        """pushes raster data to object, you can also use a value, rasters should be aligned"""

        if raster.shape != self.shape:
            raise ValueError(
                "Input is not aligned! Use input_raster = input_raster.align(raster)"
            )

        target = self.copy(shell=True)
        self.optimize_blocksize = False
        raster.optimize_blocksize = False

        logger.debug("raster - push_raster - start")
        for target_tile, raster_tile in zip(self, raster):
            target_array = target_tile.array
            raster_mask = raster_tile.mask

            if value:
                target_array[raster_mask] = value
            else:
                raster_array = raster_tile.array
                target_array[raster_mask] = raster_array[raster_mask]
            target.array = (target_array, *raster_tile.location)

        del target_array
        del raster_mask
        return target

    def polygonize(
        self,
        field_name="dn",
        layer_name="poly",
        mask_value=None,
        idx=0,
        no_mask=False,
        quiet=True,
    ):
        """polygonizes a raster
        params:
            field_name: field in vector
            layer_name: layer in vector
            mask_value: a value will be set to nodata
            no_mask: also polygonizes nodata
            quiet: if false it shows progress

        return:
            path: str

        """
        polygonize_type = TYPE_MAPPING[self.band.DataType]
        path = loc_dir(ext="shp", mem=True)
        ds_vector = DRIVER_OGR_SHP.CreateDataSource(path)
        layer_vector = ds_vector.CreateLayer(layer_name, srs=self.spatial_reference)
        raster_field = ogr.FieldDefn(field_name, polygonize_type)
        layer_vector.CreateField(raster_field)

        if mask_value is not None:
            logger.debug("loading raster in memory due to mask value edits")
            self.load_to_memory()
            array = self.array
            array[array == mask_value] = self.nodata_value
            self.array = array
            no_mask = False

        band = self.band
        array = self.array

        if no_mask:
            no_data_band = None
        else:
            no_data_band = band.GetMaskBand()

        band.FlushCache()

        if not quiet:

            progress = Progress(total=100, message="Polygonize", quiet=quiet)
            callback = progress.gdal
        else:
            callback = None

        if polygonize_type == ogr.OFTReal:
            gdal.FPolygonize(
                band,
                no_data_band,
                layer_vector,
                idx,
                [],
                callback=callback,
            )
        else:
            gdal.Polygonize(
                band,
                no_data_band,
                layer_vector,
                idx,
                [],
                callback=callback,
            )

        return path

    def align(self, template=None, fill_value=None, idw=False, nodata_align=True):
        return align(self, template, fill_value, idw, nodata_align)

    def fill_nodata(
        self,
        search_distance=10,
        smooth_iterations=0,
        max_iterations=10,
        array=None,
        quiet=False,
    ):

        """
        Inverse distance weighting - Fills a raster until all nodata is filled.
        For each iterations one pixel if filled
        Params:
            Search Distance: Maximum distance (in pixels) to search out for values to interpolate
                             This is also the amount of nodata pixels it will fill
            Smooth iterations: Number of smoothing iterations to run after the interpolation
            Max Iterations: Limits the amounts of IDW loops


        return a Raster object
        """

        if array is None:
            array = self.array  # nan
        else:
            array = self.nan(array)

        # back to  not np.nan values
        array[np.isnan(array)] = self.nodata_value

        kwargs = self.settings
        kwargs["spatial_reference"] = self.spatial_reference.wkt

        # combine both nans and nodata_values
        fill_mask = array == self.nodata_value
        nan_count = np.count_nonzero(array[fill_mask])

        target = self.copy(shell=True)
        idw_iterations = 0
        while nan_count > 0:
            idw_iterations += 1
            if not quiet:
                logger.debug(f"IDW iteration {idw_iterations}, nan count {nan_count}")

            with Dataset(array, **kwargs) as work:
                work_band = work.GetRasterBand(1)
                mask_band = work_band.GetMaskBand()

                gdal.FillNodata(
                    work_band,
                    mask_band,
                    search_distance,  # search distance
                    smooth_iterations,  # smoothing iterations
                )

            nan_count = np.nansum(array == self.nodata_value)

            if idw_iterations >= max_iterations:
                logger.debug(f"Reached max iterations {max_iterations}")
                break

        if idw_iterations < max_iterations:
            logger.debug("No more nodata left")

        target.array = array
        return target

    def read_point(self, point: ogr.Geometry, nodata_value=-9999):
        """
        Samples points, based on 'raster'.
        input is an ogr geometry
        return a dictionary with sampled values
        """
        # inverse
        p, a, b, q, c, d = self.geotransform
        D = 1 / (a * d - b * c)
        e, f, g, h = d * D, -b * D, -c * D, a * D

        x, y = point.GetX(), point.GetY()
        j = np.int64(e * (x - p) + f * (y - q))
        i = np.int64(g * (x - p) + h * (y - q))

        try:
            array = self.band.ReadAsArray(int(j), int(i), 1, 1).astype(np.float)
            return array[0][0]
        except:
            return nodata_value  # Point outside raster?

    def read_geometry(self, geometry: ogr.Geometry, count=False):
        """
        Return numpy array.

        :param inflate: inflate envelope to grid, to make sure that
            the entire geometry is contained in resulting indices.

        If the bounds fall outside the dataset, the result is padded
        with no data values.

        This ingenuity is from raster-tools
        """

        geotransfrom = self.geotransform
        spatial_reference = self.spatial_reference
        nodata_value = self.nodata_value

        x1, y1, x2, y2 = geotransfrom.get_indices(geometry, inflate=True)

        # overlapping bounds
        w, h = self.columns, self.rows
        p1 = min(w, max(0, x1))
        q1 = min(h, max(0, y1))
        p2 = min(w, max(0, x2))
        q2 = min(h, max(0, y2))

        # result array plus a view for what's actually inside datasets
        array = np.full((y2 - y1, x2 - x1), nodata_value, self.np_data_type)
        view = array[q1 - y1 : q2 - y1, p1 - x1 : p2 - x1]

        kwargs = {
            "xoff": p1,
            "yoff": q1,
            "win_xsize": p2 - p1,
            "win_ysize": q2 - q1,
        }
        data = self.band.ReadAsArray(**kwargs)
        index = data != nodata_value
        view[index] = data[index]

        if 0 in view.shape:
            if count:
                return view, 1
            return view

        # not using self.settings due to speed
        kwargs = {
            "spatial_reference": spatial_reference.wkt,
            "geotransform": geotransfrom.shifted(geometry, inflate=True),
            "nodata_value": nodata_value,
        }

        select_2d = np.zeros(view.shape, dtype="u1")
        with Layer(geometry) as layer:
            with Dataset(select_2d[np.newaxis], **kwargs) as dataset:
                gdal.RasterizeLayer(dataset, [1], layer, burn_values=[1])

        if array.shape != select_2d.shape:
            if count:
                return np.array([0]), 1
            return np.array([0])

        else:
            array_1d = array[select_2d.astype("b1")]

        # determine data or no data
        if array_1d.dtype.kind == "f":
            select_1d = ~np.isclose(array_1d, nodata_value)
        else:
            select_1d = ~np.equal(array_1d, nodata_value)

        if count:
            return array_1d[select_1d], len(array_1d)

        return array_1d[select_1d]

    def read(self, geometry: ogr.Geometry, count=False):

        """
        Return numpy array or a value
        Functions combines an read_geometry and read_point
        params:
            count: If True, returns the amount of selected pixels including nodata.

        """
        # find indices
        if geometry.GetGeometryType() == ogr.wkbPoint:
            return self.read_point(geometry)
        else:
            return self.read_geometry(geometry, count)

    def classify(self, new_classes, quiet=True):
        """
        classifies the raster based on a dictionary.
        e.g.,
            new_classes = {raster_value: [new_value]}
        """
        logger.info("start classifying raster")

        target = self.copy(shell=True, data_type=gdal.GDT_Float32)
        for tile in Progress(self, "Classifying", quiet):
            array = tile.array

            if np.isnan(array).all():
                target.array = array, *tile.location
                continue

            codes = np.unique(array[~np.isnan(array)])

            output_array = np.copy(array)
            for code in codes:
                output_array[array == code] = new_classes[code]

            target.array = output_array, *tile.location

            # # Delete array
            del array
            del output_array

        return target

    def write(self, filename, overwrite=True, write_options=[]):
        filename = str(filename)  # pathlib
        if overwrite:
            if os.path.exists(filename):
                DRIVER_GDAL_GTIFF.Delete(filename)

        WRITE_OPTIONS.extend(write_options)
        ds = DRIVER_GDAL_GTIFF.CreateCopy(filename, self.ds, options=WRITE_OPTIONS)
        ds.GetRasterBand(1).SetNoDataValue(self.nodata_value)
        ds.SetProjection(self.spatial_reference.wkt)
        ds.SetGeoTransform(self.geotransform)
        ds.SetMetadata(self.metadata)
        ds = None

        return


def loc_dir(_dir=None, _type="ras", ext="tif", mem=False, add=True):
    """Creates a string to be used in a datasource"""
    global _loc_ras
    if add:
        _loc_ras = _loc_ras + 1

    if mem:
        _dir = "/vsimem"
    else:
        _dir = tempfile.mkdtemp()

    return _dir + f"/{_type}_{_loc_ras}.{ext}"


def create_source(columns, rows, data_type, memory=True):
    """Takes columns, rows and a data_type to create an in memory gdal datasource"""
    path = loc_dir(mem=memory, add=True)
    if memory:
        driver = DRIVER_GDAL_MEM
    else:
        driver = DRIVER_GDAL_GTIFF

    out = driver.Create(
        path,
        columns,
        rows,
        1,
        data_type,
        options=WRITE_OPTIONS,
    )
    return out, path


def moving_window_generator(
    array, row_min=1, row_max=1, column_min=1, column_max=1, quiet=True
):
    """generates a moving window on top of the class"""
    rows, columns = array.shape
    for row in range(rows):
        for column in range(columns):

            # level 2 create window
            bottom = row - row_min
            top = row + row_max
            left = column - column_min
            right = column + column_max

            # window check
            if bottom < 0:
                bottom = 0
            if top > rows:
                top = rows
            if left < 0:
                left = 0
            if right > columns:
                right = columns
            # extract views
            yield array[bottom : top + 1, left : right + 1], (
                bottom,
                top,
                left,
                right,
            ), (row, column)


def optimize_blocksize(dataset, target_pixels=1000000):
    """Takes a band and creates an optimzed blocksize
    1. Blocksizes are at least orginal band block sizes
    2. New blocksizes are a multitude of original sizes
    3. Block size cannot be larger than the raster size
    4. If target pixels are approximated.

    """

    band = dataset.GetRasterBand(1)
    x_blocksize, y_blocksize = band.GetBlockSize()
    total_block_pixels = x_blocksize * y_blocksize
    total_raster_pixels = dataset.RasterXSize * dataset.RasterYSize

    x_optimized = x_blocksize
    y_optimized = y_blocksize

    while total_block_pixels < target_pixels:
        pre_x_optimized = x_optimized
        pre_y_optimized = y_optimized

        # block gets larger with multitude of blocks
        if x_optimized < y_optimized:
            x_optimized += x_blocksize
        else:
            y_optimized += y_blocksize

        # block cannot be larger than total_raster
        total_block_pixels = x_optimized * y_optimized
        if total_block_pixels > total_raster_pixels:
            x_optimized = pre_x_optimized
            y_optimized = pre_y_optimized
            break

    return x_optimized, y_optimized


def tile_list(dataset: gdal.Dataset, optimize=True, blocksize_x=2048, blocksize_y=2048):
    """Takes a gdal ds and returns a list of coords based on blocks"""
    if optimize:
        x_blocksize, y_blocksize = optimize_blocksize(dataset)
    else:
        x_blocksize = blocksize_x
        y_blocksize = blocksize_y

    x_size = dataset.RasterXSize
    y_size = dataset.RasterYSize

    tiles = []
    for x in range(0, x_size, x_blocksize):

        if x + x_blocksize < x_size:
            cols = x_blocksize
        else:
            cols = x_size - x

        for y in range(0, y_size, y_blocksize):

            if y + y_blocksize < y_size:
                rows = y_blocksize
            else:
                rows = y_size - y
            tiles.append((x, y, cols, rows))
    return tiles


def tile_generator(
    dataset: gdal.Dataset, optimize: bool = True, blocksize_x=2048, blocksize_y=2048
):
    """returns a list of tiles in memory"""
    band = dataset.GetRasterBand(1)
    transform = dataset.GetGeoTransform()
    nodata_value = band.GetNoDataValue()
    projection = dataset.GetProjection()

    for x, y, cols, rows in tile_list(dataset, optimize, blocksize_x, blocksize_y):
        tile_array = band.ReadAsArray(x, y, cols, rows)
        minx = transform[0] + transform[1] * x
        maxy = transform[3] + transform[5] * y
        tile_gt = (minx, transform[1], 0, maxy, 0, transform[5])
        tile_ds, path = create_source(cols, rows, gdal.GDT_Float32)
        tile_ds.SetGeoTransform(tile_gt)
        tile_ds.SetProjection(projection)
        tile_ds.GetRasterBand(1).SetNoDataValue(nodata_value)
        tile_ds.GetRasterBand(1).WriteArray(tile_array)
        tile_array = None
        yield tile_ds, (x, y)


def align(raster, template=None, fill_value=None, idw=False, nodata_align=True):
    """
    Takes a Raster object and returns a aligned raster.
    Assumes the current location of the rasters is correct
    and a similar pixel size
    params:
        template: raster to which the raster is aligned
        fill_value: if nodata needs to be filled, it will use this fill value
        idw: Uses idw to fill nodata needs
        nodata_align: If checked, nodata is aligned

    """

    # template info
    gtt = template.geotransform
    columns = template.columns
    rows = template.rows

    # coordinate difference
    gti = raster.geotransform
    gtt_x, gtt_y = (gtt[0], gtt[3])
    gti_x, gti_y = (gti[0], gti[3])
    x_dif = (gtt_x - gti_x) / gtt[1]
    y_dif = (gtt_y - gti_y) / abs(gtt[5])

    logger.debug("Padding and deleting raster pixels where needed")

    raster_array = raster.array
    raster_array[np.isnan(raster_array)] = raster.nodata_value

    if x_dif > 0:
        raster_array = raster_array[:, abs(int(x_dif)) :]
        # raster is to much to the left, delete values to the left
    elif x_dif < 0:
        # raster is too much to the right, add values on the left side
        raster_array = np.pad(
            raster_array,
            (
                (0, 0),  # pad upside  # pad downside
                (abs(int(x_dif)), 0),  # pad left side  # pad on the right side
            ),
            constant_values=raster.nodata_value,
        )

    col_dif = columns - raster_array.shape[1]
    if col_dif > 0:
        # add columns, width is too small
        raster_array = np.pad(
            raster_array,
            [
                (0, 0),  # pad upside  # pad downside
                (
                    0,
                    abs(int(col_dif)),
                ),  # pad left side  # pad on the right side
            ],
            constant_values=raster.nodata_value,
        )

    elif col_dif < 0:
        # delete columns
        raster_array = raster_array[:, 0 : raster_array.shape[1] - abs(int(col_dif))]

    # works
    if y_dif > 0:
        # raster is to low
        # add rows on the upside if raster is to low
        raster_array = np.pad(
            raster_array,
            [
                (abs(int(y_dif)), 0),  # pad upside  # pad downside
                (0, 0),  # pad left side  # pad on the right side
            ],
            constant_values=raster.nodata_value,
        )
    elif y_dif < 0:
        # raster is too high
        # delete rows from upside if raster is too hgh
        raster_array = raster_array[abs(int(y_dif)) :, :]

    row_dif = rows - raster_array.shape[0]
    if row_dif > 0:
        raster_array = np.pad(
            raster_array,
            [
                (0, abs(int(row_dif))),  # pad upside  # pad downside
                (0, 0),  # pad left side  # pad on the right side
            ],
            constant_values=raster.nodata_value,
        )

    elif row_dif < 0:
        # raster is too large in height
        raster_array = raster_array[0 : raster_array.shape[0] - abs(int(row_dif)) :, :]

    if nodata_align == True:
        logger.debug("Padding nodata values onto the to be aligned array")

        nodata_index = np.isnan(template.array)
        raster_array[nodata_index] = raster.nodata_value

        if np.sum(raster_array == raster.nodata_value) != np.sum(nodata_index):
            if idw == True:
                pass  # done later in the process

            elif not fill_value == None:
                (values, counts) = np.unique(raster_array, return_counts=True)
                fill_value = values[np.argmax(counts)]
                if fill_value == raster.nodata_value:
                    fill_value = values[np.argmax(counts) - 1]

                raster_array[
                    ~nodata_index & (raster_array == raster.nodata_value)
                ] = fill_value

                logger.debug(
                    f"Padding fill value {fill_value} onto the to be aligned array"
                )

    raster_array[raster_array == raster.nodata_value] = np.nan
    output = template.copy(shell=True)
    output.array = raster_array

    if idw == True:
        logger.debug("Using inverse distance weighting to fill up nodata")
        output.fill_nodata()

    output.name = raster.name

    return output


if HAS_SCIPY:

    def perona_malik_filter(data, iter=5, K=5, delta=0.2, method=2, **kwargs):
        """returns a filtered array"""

        dxW = np.array([[0, 0, 0], [1, -1, 0], [0, 0, 0]])
        dxE = np.array([[0, 0, 0], [0, -1, 1], [0, 0, 0]])
        dyN = np.array([[0, 1, 0], [0, -1, 0], [0, 0, 0]])
        dyS = np.array([[0, 0, 0], [0, -1, 0], [0, 1, 0]])

        pm_lambda = np.ones(data.shape) * delta

        k1 = np.ones(data.shape) * (-1.0 / K)
        k2 = np.ones(data.shape) * K * K

        for i in range(iter):

            dI_W = scipy.ndimage.filters.convolve(data, dxW)
            dI_E = scipy.ndimage.filters.convolve(data, dxE)
            dI_N = scipy.ndimage.filters.convolve(data, dyN)
            dI_S = scipy.ndimage.filters.convolve(data, dyS)

            if method == 1:

                cW = np.exp(dI_W * dI_W * k1)
                cE = np.exp(dI_E * dI_E * k1)
                cN = np.exp(dI_N * dI_N * k1)
                cS = np.exp(dI_S * dI_S * k1)

            elif method == 2:

                cW = np.ones(data.shape) / (np.ones(data.shape) + (dI_W * dI_W / k2))
                cE = np.ones(data.shape) / (np.ones(data.shape) + (dI_E * dI_E / k2))
                cN = np.ones(data.shape) / (np.ones(data.shape) + (dI_N * dI_N / k2))
                cS = np.ones(data.shape) / (np.ones(data.shape) + (dI_S * dI_S / k2))

            data = data + (pm_lambda * (cN * dI_N + cS * dI_S + cE * dI_E + cW * dI_W))

        return data



# -*- coding: utf-8 -*-
"""
Created on Wed May 25 15:37:33 2022

@author: chris

A simple version of editing a threedimodel in python
"""

# variables
import threedi_edits as tre

sqlite_path = r"C:\Users\chris.kerklaan\Documents\Projecten\overijssels_kanaal\processing\model\winter.sqlite"
point_path = r"C:\Users\chris.kerklaan\Documents\Projecten\overijssels_kanaal\processing\point.shp"
output = r"C:\Users\chris.kerklaan\Documents\Projecten\overijssels_kanaal\processing"

# First show all logging

# Load in memory so it is editable.
model = tre.ThreediEdits(sqlite_path, mode="memory")

# resettings all friction values of the cross section
cross_section_locations = model.cross_section_locations
for cross_section_location in cross_section_locations:
    cross_section_location["friction_value"] = 0.1

# show the edited part values
print(cross_section_locations[cross_section_locations.fids[0]])

# add an vertice to the channel
# get the data
points = tre.Vector(point_path)
point_geometry = points[0].geometry

# get the right channel
channels = model.channels

# get the channel by filter
channel = channels.filter(code="18049")[0]

# get the channel by id
channel = channels[26064]

# see what you can do with the geometry
dir(channel.geometry)
help(channel.geometry.add_vertice)

# show differences in linestring:
# Keep snapping size as low as possible
print(channel.geometry.add_vertice(point_geometry))
print(channel.geometry)

channel.geometry = channel.geometry.add_vertice(point_geometry)

# Templates are used to add a row.
templates = tre.Templates()
cross_section = templates.cross_section

# You can look up some of the documentation by printing the template
print(cross_section)

# Edit the template
cross_section["height"] = 5

# print again to show
print(cross_section)

# add to the cross sections
cross_section_defintions = model.cross_section_definitions

# Note this gives an error, because code cannot be None
cross_section_defintions.add(cross_section)

# Edit all other things
cross_section["code"] = "x"
cross_section["shape"] = 1
cross_section["width"] = 1

# Add
cross_section_defintions.add(cross_section)

# You can write and check the whole model at once
model.write(output + "/model2.sqlite", rasters=True)

# or write only the cross sections for a faster look
cross_section_locations.write(output + "/csl.gpkg")
channels.write(output + "/channel.gpkg")

raster = model.rasters

raster.check_properties()

dem = raster.dem
infiltratie = raster.infiltration

output = dem.copy()
for tile in dem:
    array = tile.array * 1
    output.array = (tile.location, array)

raster.dem = output



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 15 08:57:13 2021

@author: chris.kerklaan
    
"""
#

# Third-party imports
import logging
from osgeo import ogr

# Local imports
from threedi_edits.gis.vector import Vector
from threedi_edits.gis.feature import Feature
from threedi_edits.threedi.feature import ThreediFeature

from threedi_edits.globals import SUPPORTED_THREEDI_VERSIONS
from threedi_edits.threedi.constants import get_version

# Globals
V = get_version(SUPPORTED_THREEDI_VERSIONS[0])

PROPERTIES = V.Properties()
ThreediTemplateTable = V.ThreediTemplateTable
logger = logging.getLogger(__name__)


class ThreediVector(Vector):
    """
    The threedivector class retrieves special threedi features
    and is able to add data via a template.
    """

    instance = "threedi.vector.ThreediVector"

    def __init__(self, ogr_ds, table_name):
        Vector.__init__(self, path=None, datasource=ogr_ds, layer_name=table_name)
        self.name = table_name
        self.properties = PROPERTIES[table_name]

        if not hasattr(self, "layer"):
            logger.info(f"Table {table_name} not present.")
            return

    def __iter__(self):
        fids = [i.GetFID() for i in self.layer]
        for fid in fids:
            yield self.get_feature(fid)

    def __getitem__(self, i):
        return self.get_feature(i)

    def __repr__(self):
        fids = self.fids
        if len(fids) > 5:
            fid_print = f"{fids[:5]} ... {fids[-5:]}"
        else:
            fid_print = f"{fids}"

        return f"({self.instance}) 3Di table: {self.name}, row ids: {fid_print}"

    def get_feature(self, i):
        """helper for retrieving features"""
        return ThreediFeature(
            self.layer.GetFeature(i),
            self.layer,
            self.fields,
            self.layer_name,
        )

    def template_data(self, template: dict):
        """input is a ThreediTemplate, output is geometry, fid, attributes
        None if not present
        """
        # dervive id from template
        fid = template.get("id")
        if fid == "None":
            fid = None
        else:
            fid = int(fid)

        # derive geometry from template
        geometry = template.get("the_geom")

        # derive content from template
        attributes = {}
        for key in template.keys():
            if key in ["id", "the_geom", "table"]:
                continue
            value = template[key]
            try:
                content = value.content
            except AttributeError:
                content = value
            attributes[key] = content

        return fid, attributes, geometry

    def add(self, feature=None, geometry=None, fid=None, **items):

        # Get the last one or set to the first.

        if isinstance(feature, ThreediTemplateTable):
            fid, items, geometry = self.template_data(feature)

        elif isinstance(feature, Feature) or isinstance(feature, ogr.Feature):
            items = feature.items
            geometry = feature.geometry

        else:
            assert len(items) != 0

        if fid is None:
            if self.fids[-1] > 0:
                fid = self.fids[-1] + 1
            else:
                fid = 1

        return self._add(geometry=geometry, fid=fid, **items)



# -*- coding: utf-8 -*-
"""
Created on Mon Feb 15 08:52:16 2021

@author: chris.kerklaan

#TODO:
    
# The current threedi_model_checker has needs an 'migrated sqlite'. However,
when you update the migrated sqlite, you increase
    
"""

# First-party imports
import shutil
import pathlib
import logging
import types

# Third-party imports
from osgeo import ogr
from osgeo import osr

# Local imports
import threedi_edits as te
from threedi_edits.gis.vectorgroup import VectorGroup
from threedi_edits.gis.vector import Vector
from threedi_edits.threedi.vector import ThreediVector
from threedi_edits.globals import SUPPORTED_THREEDI_VERSIONS
from threedi_edits.threedi.constants import get_version

# Globals
FILE_PATH = pathlib.Path(__file__)
EMPTY_SQLITE_PATH = str(
    FILE_PATH.parent / "data" / SUPPORTED_THREEDI_VERSIONS[0] / "empty_klondike.sqlite"
)
V = get_version(SUPPORTED_THREEDI_VERSIONS[0])

# Drivers
OGR_SQLITE_DRIVER = ogr.GetDriverByName("SQLite")

# Logger
logger = logging.getLogger(__name__)


class ThreediVectorGroup(VectorGroup):
    """
    ThreediVectorGroup's makes it easier to read, write and copy threedi-objectes.
    params:
        ogr_ds: ogr.DataSource
        mode:
            'write': Used to write a file
            'empty': Returns an empty in-mem ogr threedi model
            'memory': The threedimodel is copied into memory, makes speedy edits
            'read': Used for reading a model only


    """

    instance = "threedi.vectorgroup.VectorGroup"

    def __init__(self, ogr_ds: ogr.DataSource, mode="memory"):

        self.ds = ogr_ds  # must be here to ensure correct copying.
        if mode == "write":
            super().__init__(ogr_ds=ogr_ds, memory=False, write=True)
        elif mode == "read":
            super().__init__(ogr_ds=ogr_ds, memory=False, write=False)
        elif mode == "empty":
            super().__init__(ogr_ds=create_empty_model().ds, memory=False)
        elif mode == "memory":
            super().__init__(ogr_ds=self._copy(ogr_ds), memory=False)

        logger.debug(f"Opened threedi model in {mode} mode")

        self.srs = osr.SpatialReference()
        self.srs.ImportFromEPSG(4326)
        self.mode = mode

    def __str__(self):
        return "3Di model"

    def __repr__(self):
        return f"({self.instance}) 3Di model"

    def __iter__(self):
        for table_name in V.TABLES["all"]:
            if table_name in self.layers:
                yield self._table(table_name)

    def _tables(self):
        tables = types.SimpleNamespace()
        for table in V.TABLES["all"]:
            if table in V.TRANSLATE:
                setattr(tables, V.TRANSLATE[table], self._table(table))

        return tables

    def _table(self, table_name):
        return ThreediVector(self.ds, table_name)

    def __getitem__(self, table):
        return self._table(table)

    def __setitem__(self, table_name, ogr_layer):
        """replaces an entire table"""
        self.ds.DeleteLayer(table_name)
        self.ds.CreateLayer(ogr_layer)

    def _copy(
        self, memory: bool = True, path: str = None, quiet=True
    ) -> ogr.DataSource:
        """Copies the current either to memory or to a path"""
        if memory:
            output = create_empty_model()
        else:
            output = create_output_model(path)

        current_layers = output.layers
        for table in self:
            # Some layers are remove, so we are skipping some things.
            if table.name not in current_layers:
                continue
            if table.name in V.TABLES["skip"]:
                continue

            table_output = output[table.name]
            subject = f"Copying {table.name} ({table.count} rows)"
            for feature in te.Progress(table, subject, quiet):
                table_output.add(
                    items=feature.evaluate_types(),
                    geometry=feature.geometry,
                    fid=feature.fid,
                )
            table_output = None

        return output.ds

    def _write(self, path, quiet=False):
        """
        Whilst copy layer is faster than iterating over every feature,
        we want check the feature on field types and use the sqlite constraints
        therefore iteration is chosen here
        """
        self._copy(False, path, quiet)


def create_output_model(path):
    """Creates an empty sqlite"""
    shutil.copyfile(EMPTY_SQLITE_PATH, path)
    output = ThreediVectorGroup(ogr.Open(path, 1), mode="write")
    empty_tables = [
        "v2_global_settings",
        "v2_aggregation_settings",
        "v2_numerical_settings",
        "schema_version",
    ]

    for empty_table in empty_tables:
        table = output[empty_table]
        table.table
        for fid in table.fids:
            table.delete(fid)
        table = None

    return output


def create_empty_model():
    """creates a ThreediModelBase from scratch"""
    group = VectorGroup.from_scratch(name="Threedi Model")
    for name, threedi_property in V.Properties():

        # create vector, with the right geometry and fields
        if "the_geom" in threedi_property:
            geometry_type = threedi_property["the_geom"]["type"]
        else:
            geometry_type = ogr.wkbUnknown

        vector = Vector.from_scratch(name, geometry_type, 4326)

        for i, (field_name, field_values) in enumerate(threedi_property.items()):
            if field_name in ["connected_tables", "the_geom", "id"]:
                continue
            vector.add_field(field_name, field_values["type"])

        # add to the group
        group.add(vector, name)

    return group
